//*****************************************************************************
//  Copyright (C) 2007 Microsoft Corporation
//  All rights reserved.
//*****************************************************************************

#include "Globals.h"
#include <iostream>
#include <fstream>
#ifdef _MSC_VER
    #include <strsafe.h>
    #include <aclapi.h>
    #include <Globals.tmh>
#else
    #include <execinfo.h>
#endif

using namespace std;
using namespace TestSystem;

PWSTR ConvertStringToW(_In_z_ const char *buf)
{
#ifdef _MSC_VER
    int size = MultiByteToWideChar(CP_THREAD_ACP, 0, buf, -1, NULL, NULL);
#else
    int size = mbstowcs(0, buf, 0);
#endif
    if (size == 0)
    {
        throw Exception();
    }



#ifdef _MSC_VER
    PWSTR widebuf = new wchar_t[size];
    size = MultiByteToWideChar(CP_THREAD_ACP, 0,
            buf, -1, widebuf, size);
#else
    PWSTR widebuf = new wchar_t[size + 1];
    size = mbstowcs(widebuf, buf, size);
#endif
    if (size == 0)
    {
        delete [] widebuf;
        throw Exception();
    }
#ifndef _MSC_VER
    else
    {
        widebuf[size] = L'\0';
    }
#endif

    return widebuf;
}

PSTR ConvertStringToA(_In_z_ const wchar_t *buf)
{
#ifdef _MSC_VER
    int size = WideCharToMultiByte(CP_THREAD_ACP, 0, buf, -1, NULL, NULL, NULL, NULL);
#else
    int size = wcstombs(NULL, buf, 0);
#endif
    if (size == 0)
    {
        throw Exception();
    }

#ifdef _MSC_VER
    PSTR ansibuf = new char[size];
    size = WideCharToMultiByte(CP_THREAD_ACP, 0, buf, -1, ansibuf, size, NULL, NULL);
#else
    PSTR ansibuf = new char[size + 1];
    size = wcstombs(ansibuf, buf, size + 1);
#endif
    if (size == 0)
    {
        delete [] ansibuf;
        throw Exception();
    }
#ifndef _MSC_VER
    else
    {
        ansibuf[size] = '\0';
    }
#endif

    return ansibuf;
}

namespace TestSystem
{

bool StartsWith(_In_ PCWSTR str, _In_ PCWSTR prefix)
{
    return wcsncmp(str, prefix, wcslen(prefix)) == 0;
}

bool Equals(_In_ PCWSTR str, _In_ PCWSTR target)
{
    return wcscmp(str, target) == 0;
}

bool EqualsCI(_In_ PCWSTR str, _In_ PCWSTR target)
{
    return PAL_wcsicmp(str, target) == 0;
}

PWSTR Copy(_In_opt_ PCWSTR str)
{
    if (str == NULL)
    {
        return NULL;
    }

    int length = wcslen(str);
    wchar_t *copy = new wchar_t[length + 1];
    if (copy == NULL)
    {
        throw Exception();
    }
    PAL_CopyByLenW(copy, str, length + 1);

    return copy;
}

} //namespace TestSystem

#ifdef _MSC_VER
static SECURITY_DESCRIPTOR g_SecurityDescriptor = {
    SECURITY_DESCRIPTOR_REVISION,
    0,
    SE_DACL_PRESENT,
    NULL,
    NULL,
    NULL,
    NULL,
};

static SECURITY_ATTRIBUTES g_SecurityAttributes = {
    sizeof(SECURITY_ATTRIBUTES),
    &g_SecurityDescriptor,
    FALSE,
};

NITS_EXPORT void NitsEventCreate(
    _Inout_ Event *event,
    _In_ PCWSTR name,
    NitsFaultMode mode)
{
    event->m_handle = NULL;
    event->m_name = name;

    if (wcslen(name) > Event::MaxName)
    {
        FatalError();
    }

    event->m_handle = CreateEvent(&g_SecurityAttributes, mode == Manual, FALSE, name);
    if (event->m_handle == NULL)
    {
        FatalError();
    }
}

NITS_EXPORT void NitsEventDelete(
    _Inout_ Event *event)
{
    if (event->m_handle)
    {
        CloseHandle(event->m_handle);
        event->m_handle = NULL;
    }
}

#endif

DWORD Mapping::Initialize(_In_reads_bytes_(bytes) void const *context, long bytes)
{
    if (m_bytes == 0)
    {
        m_bytes = bytes;

        m_copy = malloc(m_bytes);

        if (m_copy == NULL)
            throw Exception();

        memcpy(m_copy, context, m_bytes);
    }
    else if (m_bytes != bytes)
    {
        return ERROR_INVALID_PARAMETER;
    }


    if (m_mapping == NULL)
    {
        m_mapping = PAL_SharedMemory_Open(m_name,
                                        m_bytes + sizeof(long));
        if (m_mapping == NULL)
        {
#ifdef _MSC_VER
            return GetLastError();
#else
            return ERROR_OUTOFMEMORY;
#endif
        }
    }

    if (m_status == NULL)
    {
        //Memory is zeroed for the first process. m_status is Unloaded.
        //In that case, m_state is Unloaded.
        m_status = reinterpret_cast<long *>(PAL_SharedMemory_Map(m_mapping, m_bytes + sizeof(long)));
        if (m_status == NULL)
        {
#ifdef _MSC_VER
            return GetLastError();
#else
            return ERROR_OUTOFMEMORY;
#endif
        }

        m_data = const_cast<long *>(m_status + 1);
        if (PAL_InterlockedCompareExchange(m_status, Loading, Unloaded) == Unloaded)
        {
            PAL_CopyMemory(m_data, context, m_bytes);
            *m_status = Loaded;
        }
    }

    while (*m_status != Loaded)
    {
        PAL_Sleep(10);
    }

    return NO_ERROR;
}

void Mapping::Reset()
{
    memcpy(m_data, m_copy, m_bytes);
}

Fault::Fault() : m_lock(0)
{
    Reset(CallSite_NONE, 0, false);

    m_mainThread = 0;
    m_minimumAttemptDifferentThread = 0;
}

void Fault::Reset(int site, int iteration, bool breakOnFault)
{
    m_site = site;
    m_iteration = iteration;
    m_attempt = 0;

    m_break = breakOnFault;
    m_faulted = false;
    m_filtered = false;
    m_hit = CallSite_NONE;
    m_line = 0;
    m_file[0] = '\0';

    m_faultedAttempt = 0;
    m_firstAttemptDifferentThread = 0;
}

bool Fault::ShouldFault(CallSite const &site)
{
    if (m_site == CallSite_NONE)
    {
        return false;
    }

    if (m_site != CallSite_ALL && m_site != site.id)
    {
        return false;
    }

    while (PAL_InterlockedExchange(&m_lock, TRUE) == TRUE)
    {
        PAL_Sleep(10);
    }

    long attempt = PAL_InterlockedIncrement(&m_attempt);
    if (m_site == CallSite_ALL && GetGlobals().GetConfiguration().mode >= NitsTestStackFault)
    {
        bool filtered = false;
        int limit = 0;
        // TODO: Fix this when actual PAL is integrated with nits.
        if (PAL_GetCurrentThreadId() != (DWORD)m_mainThread)
        {
            //Faults from other threads are more tightly controlled.
            //Attempts earlier than the given limit should be filtered out.
            limit = m_minimumAttemptDifferentThread;
            if (m_firstAttemptDifferentThread == 0)
            {
                m_firstAttemptDifferentThread = attempt;
            }
        }
        if (m_faulted || !GetGlobals().m_frames.ShouldFault(limit, attempt, filtered))
        {
            if (filtered)
            {
                m_filtered = true;
            }
            PAL_InterlockedExchange(&m_lock, FALSE);
            return false;
        }
    }
    else
    {
        if (m_iteration != 0 && m_iteration != attempt)
        {
            PAL_InterlockedExchange(&m_lock, FALSE);
            return false;
        }
    }

    m_faulted = true;
    PAL_InterlockedExchange(&m_lock, FALSE);

    NitsTraceEx(L"Simulating fault", site, NitsAutomatic);

#ifdef _MSC_VER
    DoTraceMessage(TestLifetime, L"Simulating fault!");
#endif
    //Update the last fault hit.
    m_hit = site.id;
    m_line = site.line;
    m_faultedAttempt = attempt;
    PCSTR temp = site.file;

#ifdef _MSC_VER
    CaptureStackBackTrace(3, 50, GetGlobals().GetStackTrace(), NULL);
#else
    backtrace(GetGlobals().GetStackTrace(), 50);
#endif
    //Truncate the beginning of the string.
    int length = strlen(temp);
    int limit = FileSize - 1;
    if (length > limit)
    {
        temp += length - limit;
    }

    PAL_CopyByLenA(m_file, temp, FileSize);

#ifdef _MSC_VER
    if (m_break)
    {
        DebugBreak();
    }
#endif
    return true;
}

CallSite Fault::GetFaultedSite() const
{
    return NitsMakeCallSite(m_hit, NULL, m_file, m_line);
}

Globals::Globals()
    : m_version(SharedMemoryVersion),
      m_unload(0),
      m_runLock(0),
      m_attachLock(0),
      m_pipeLock(FALSE),
      m_pipeChars(0)
{
    m_pipe[0] = '\0';
    m_debugger[0] = '\0';
    m_binaryFilter[0] = '\0';
    m_binaryTarget[0] = '\0';

    for (int i = 0; i < InjectorListSize; i++)
        m_injectors[i].process = 0;

    Reset();

    for (int i = 0; i < ResultCount; i++)
    {
        m_statistics[i] = 0;
    }
}

Globals::~Globals()
{
}

bool Globals::StartRun(bool force)
{
    m_version = SharedMemoryVersion;

    CheckRun();

    DWORD pid = PAL_GetCurrentProcessId();
    if (force)
    {
        PAL_InterlockedExchange(&m_runLock, pid);
    }
    else if (PAL_InterlockedCompareExchange(&m_runLock, pid, 0) != 0)
    {
        return false;
    }

    for (int i = 0; i < ResultCount; i++)
    {
        m_statistics[i] = 0;
    }

    m_frames.Reset();

    PAL_ZeroMemory(&m_stats, sizeof(m_stats));

    // injection is disabled on linux for now
#ifdef _MSC_VER
    //Enable all privileges to capture SeDebugPrivilege.
    HANDLE token;
    OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &token);

    DWORD length = 0;
    TOKEN_USER tu;
    GetTokenInformation(token, TokenPrivileges, &tu, sizeof(TOKEN_USER), &length);

#pragma prefast(push)
#pragma prefast(disable: 26000)
    char *buffer = (char *)malloc(length);
    char *previous = (char *)malloc(length);
    GetTokenInformation(token, TokenPrivileges, buffer, length, &length);

    memcpy(previous, buffer, length);
    TOKEN_PRIVILEGES *list = (TOKEN_PRIVILEGES *)buffer;
    for (unsigned i = 0; i < list->PrivilegeCount; i++)
        list->Privileges[i].Attributes |= SE_PRIVILEGE_ENABLED;

    AdjustTokenPrivileges(token, FALSE, list, 0, NULL, NULL);
 
    for (int i = 0; i < InjectorListSize; i++)
    {
        InjectorTarget *target = m_injectors + i;
        if (target->process == 0)
            continue;

        HANDLE p = OpenProcess(PROCESS_DUP_HANDLE, FALSE, target->process);
        if (p == NULL)
        {
            /* This process is not accessible. Remove the entry. */
            target->process = 0;
            continue;
        }
        
        HANDLE sem;
        if (!DuplicateHandle(p, target->signalSemaphore, GetCurrentProcess(), &sem, 0, FALSE, DUPLICATE_SAME_ACCESS))
        {
            target->process = 0;
            continue;
        }

        /* Kick the signalSemaphore for each injector. */
        ReleaseSemaphore(sem, 1, NULL);
        CloseHandle(sem);
        CloseHandle(p);
    }

    for (int i = 0; i < InjectorListSize; i++)
    {
        InjectorTarget *target = m_injectors + i;
        if (target->process == 0)
            continue;

        HANDLE p = OpenProcess(PROCESS_DUP_HANDLE, FALSE, target->process);
        if (p == NULL)
        {
            /* This process is not accessible. Remove the entry. */
            target->process = 0;
            continue;
        }
        
        HANDLE sem;
        if (!DuplicateHandle(p, target->waitSemaphore, GetCurrentProcess(), &sem, 0, FALSE, DUPLICATE_SAME_ACCESS))
        {
            target->process = 0;
            continue;
        }

        /* Wait on the waitSemaphore for each injector. */
        WaitForSingleObject(sem, 100);
        CloseHandle(sem);
        CloseHandle(p);
    }

    //Restore privileges back to the way they were.
    free(buffer);
    free(previous);
    list = (TOKEN_PRIVILEGES *)previous;
    AdjustTokenPrivileges(token, FALSE, list, 0, NULL, NULL);
    CloseHandle(token);
#pragma prefast(pop)
#endif
    return true;

}

void Globals::StopRun()
{
    m_runLock = 0;
}

void Globals::CheckRun()
{
    if ((m_runLock == 0) || (m_runLock == (long)(PAL_GetCurrentProcessId())) )
    {
        return;
    }

#ifdef _MSC_VER
    HANDLE h = OpenProcess(SYNCHRONIZE, FALSE, m_runLock);
    if (h == NULL)
    {
        DWORD err = GetLastError();
        //Access denied likely means some test didn't call RevertToSelf!
        if (err != ERROR_ACCESS_DENIED)
        {
            wcerr << L"WARNING: Previous test process (PID " << m_runLock << L") exited abnormally!\n";
            StopRun();
        }
    }
    else
    {
        CloseHandle(h);
    }
#endif
}

void Globals::LockPipe()
{
    while (PAL_InterlockedExchange(&m_pipeLock, TRUE) == TRUE)
    {
        PAL_Sleep(10);
    }
}

void Globals::UnlockPipe()
{
    m_pipeLock = FALSE;
}

void Globals::PostPipe(_In_ PCWSTR text)
{
    CheckRun();

    PAL_Semaphore *event = System::GetInstance().GetPipeEvent();

    size_t length = wcslen(text);
    wchar_t *end = m_pipe + PipeSize - 1;

    while (true)
    {
        if (!m_runLock)
        {
            //Don't post unless there is someone reading.
            return;
        }

        LockPipe();

        //Append as much text as we can to the pipe.
        wchar_t *buffer = m_pipe + m_pipeChars;
        size_t available = end - buffer;
        size_t toUse = min(length, available);

        PAL_CopyMemory(buffer, text, toUse * sizeof(wchar_t));
        m_pipeChars += toUse;
        m_pipe[m_pipeChars] = '\0';

        UnlockPipe();


        //Wake up the pipe thread.
        PAL_SignalSemaphore(event);

#ifdef _MSC_VER
        if (IsDebuggerPresent())
        {
            //Issuing interactive debugger commands starves the pipe thread.
            //Make this synchronous in that case.
            while (m_pipeChars > 0)
            {
                PAL_Sleep(10);
            }
        }
#endif
        if (length == toUse)
        {
            return;
        }

        text += toUse;
        length -= toUse;
        PAL_Sleep(10);
    }
}

void Globals::DumpPipe()
{
    LockPipe();

    static ofstream s_log("nits.log");
    char buf[PipeSize];
#ifdef _MSC_VER
    WideCharToMultiByte(CP_THREAD_ACP, 0, 
            m_pipe, wcslen(m_pipe) + 1, buf, sizeof(buf), NULL, NULL);
#else
    int convertCount = wcstombs(buf, m_pipe, PipeSize);
    buf[convertCount] = '\0';
#endif

    s_log << buf;

#ifdef _MSC_VER
    wcerr << m_pipe;
    OutputDebugString(m_pipe);
#else
    wstring toPrint(m_pipe);
    wcout << toPrint;
#endif

    m_pipeChars = 0;
    m_pipe[0] = '\0';

    UnlockPipe();

#ifdef _MSC_VER
    //See if another process wants a debugger attached.
    if (m_attachLock)
    {
        wostringstream buf;
        buf << m_debugger << L" -p " << m_attachLock;

        STARTUPINFO si;
        PROCESS_INFORMATION pi;

        PAL_ZeroMemory(&si, sizeof(si));
        si.cb = sizeof(si);
        PAL_ZeroMemory(&pi, sizeof(pi));

        si.dwFlags |= STARTF_USESHOWWINDOW;
        si.wShowWindow = SW_SHOWMINNOACTIVE;

        PWSTR temp = Copy(buf.str().c_str());
        if (!CreateProcess(NULL, temp, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
        {
            wcerr << L"ERROR: Failed to attach debugger! Error=" << GetLastError() << L"\n";
            GetGlobals().SetResult(Error);
        }
        else
        {
            CloseHandle(pi.hThread);
            CloseHandle(pi.hProcess);
        }
        delete [] temp;

        m_attachLock = 0;
    }
#endif
}

void Globals::CheckDebugger()
{
#ifdef _MSC_VER
    if (Equals(m_debugger, L"") || IsDebuggerPresent())
    {
        return;
    }

    //The debugger needs to be attached.
    //This will happen on the pipe thread of the master nits.exe process.
    CheckRun();
    if (!m_runLock)
    {
        //The pipe thread is not there!
        return;
    }

    DWORD pid = PAL_GetCurrentProcessId();
    while (PAL_InterlockedCompareExchange(&m_attachLock, pid, 0) != 0)
    {
        PAL_Sleep(10);
    }

    //The pipe thread clears m_attachLock once the command is executed.
    //Wait for the debugger to attach.
    while (!IsDebuggerPresent())
    {
        PAL_Sleep(10);
    }
#endif
}

void Globals::Reset()
{
    m_result = Skipped;

    m_stopReportingIgnoredErrors = false;

    m_config.traces = NitsTraceAllTests;
    m_config.mode = NitsTestCompoundFault;
    m_config.breakFault = false;
    m_config.breakAssert = false;
    m_config.skipFlakyTests = false;

    m_simAuto.Reset(CallSite_NONE, 0, false);
    m_simManual.Reset(CallSite_NONE, 0, false);

#ifdef _MSC_VER
    m_faultError = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
#else
    m_faultError = ERROR_OUTOFMEMORY;
#endif
    m_faultEvent[0] = L'\0';
    m_debugger[0] = L'\0';
}

void Globals::SetFault(int site,
                       int attempt,
                       HRESULT error,
                       _In_ PCWSTR event)
{
    m_simManual.Reset(site, attempt, false);
    m_faultError = error;
    if (!PAL_CopyByLenW(m_faultEvent, event, Event::MaxName))
    {
        throw Exception();
    }
}

void Globals::SetDebugger(_In_ PCWSTR debugger)
{
    if (!PAL_CopyByLenW(m_debugger, debugger, 1024))
    {
        throw Exception();
    }
}

//Returns 'true' if caller should return.
//Caller always returns the value of 'test'.
NitsResult Globals::TestExpression(bool test, NitsFaultMode mode)
{
    if (this == NULL)
    {
        //For asserts tested before or during setup of this structure.
        return NitsTrue;
    }

    if (test)
    {
        //Assertion succeeded. Update state and return.
        if (m_result == Skipped)
        {
            m_result = Passed;
        }
        return NitsTrue;
    }

    if (mode == NitsManual)
    {
        //This was not an expected fault simulation failure.
        //Fails the test even if running fault sim loops.
        m_result = Error;
    }
    else if (m_result != Error)
    {
        //Don't overwrite Error with Failed.
        m_result = Failed;
    }

    if (mode == Automatic && m_simAuto.DidFault())
    {
        //Asserts are expected after automatic faults.
        return NitsTrue;
    }

#ifdef _MSC_VER
    if (m_config.breakAssert)
    {
        DebugBreak();
    }
#endif
    //Determine if failed asserts are being shown.
    return m_config.traces >= NitsTraceAsserts ? NitsFalse : NitsTrue;
}

NitsResult Globals::ShouldFault(CallSite const &site, NitsFaultMode mode)
{
    CheckDebugger();

    if (m_runLock == 0)
    {
        return NitsFalse;
    }

    GetStats().shouldFaults++;

    //if (m_runLock != PAL_GetCurrentProcessId())
    //{
    //    GetStats().shouldFaultIPCs++;
    //}

    //Check for manual faults first.
    if (m_simManual.ShouldFault(site))
    {
        // TODO: this should be replaced by PAL equivalent but this can wait
#ifdef _MSC_VER
        //Check for simulated thread delay.
        if (m_faultEvent[0] != '\0')
        {
            Event event(m_faultEvent);
            WaitForSingleObject(event, INFINITE);
        }
#endif
        return NitsTrue;
    }

    return (mode == Automatic && m_simAuto.ShouldFault(site)) ? NitsTrue : NitsFalse;
}

NitsResult Globals::DidFault()
{
    return m_simAuto.DidFault() ? NitsTrue : NitsFalse;
}

bool Globals::IsUnittestRunning()
{
    return m_runLock != 0;
}

static Globals g_tempGlobals;

System::System()
    : m_globals(NULL)
{
    g_tempGlobals.m_version = SharedMemoryVersion;

    if (NO_ERROR != Initialize(L"Global\\NitsGlobalData", &g_tempGlobals, reinterpret_cast<void **>(&m_globals), sizeof(Globals)))
    {
        wcerr << L"ERROR: Failed to load shared memory. Please run NITS elevated!\n";
        throw Exception();
    }

    m_pipeEvent = PAL_CreateSemaphore("Global\\NitsPipeEvent");
    if (m_pipeEvent == NULL)
    {
        throw Exception();
    }

    GetGlobals().CheckDebugger();
}

System::~System()
{
    PAL_DestructSemaphore(m_pipeEvent);


    for (DWORD i = 0; i < m_mappings.size(); i++)
    {
        Mapping &p = m_mappings[i];
        delete [] p.m_name;
        PAL_SharedMemory_Unmap(const_cast<long *>(p.m_status));
        PAL_SharedMemory_Close(p.m_mapping);
    }
}


Mapping &System::GetMapping(_In_ PCWSTR name)
{
    int size = m_mappings.size();
    for (int i = 0; i < size; i++)
    {
        if (Equals(m_mappings[i].m_name, name))
        {
            return m_mappings[i];
        }
    }

    Mapping temp = {Copy(name), NULL, NULL, NULL, 0};
    m_mappings.push_back(temp);
    return m_mappings[size];
}

void System::ResetClientMappings()
{
    int size = m_mappings.size();
    for (int i = 0; i < size; i++)
    {
        if (!Equals(m_mappings[i].m_name, L"Global\\NitsGlobalData"))
        {
            m_mappings[i].Reset();
        }
    }
}

Globals const &System::GetGlobals() const
{
    Globals *nothing = NULL;
    if (m_globals == NULL)
    {
        //This happens inside asserts in the SharedMemory helper.
        //The globals haven't been set up yet.
        //The assertion code avoids dereferencing the globals in this case.
#ifdef _MSC_VER
#pragma prefast(push)
#pragma prefast(disable:6011)
#endif
        return *nothing;
#ifdef _MSC_VER
#pragma prefast(pop)
#endif
    }

    return *m_globals;
}

Globals &System::GetGlobals()
{
    Globals *nothing = NULL;
    if (m_globals == NULL)
    {
        //This happens inside asserts in the SharedMemory helper.
        //The globals haven't been set up yet.
        //The assertion code avoids dereferencing the globals in this case.
#ifdef _MSC_VER
#pragma prefast(push)
#pragma prefast(disable:6011)
#endif
        return *nothing;
#ifdef _MSC_VER
#pragma prefast(pop)
#endif
    }

    return *m_globals;
}
