//*****************************************************************************
//  Copyright (C) 2008 Microsoft Corporation
//  All rights reserved.
//*****************************************************************************
#pragma once

#include "Globals.h"

#ifdef _MSC_VER
    class CWTTLogger;
#endif

namespace TestSystem
{

using namespace std;

class Param
{
public:
    Param(_In_ PCWSTR name, _In_ PCWSTR value);
    ~Param();

    PCWSTR m_name;
    PCWSTR m_value;
};

class Module
{
public:
    Module(_In_ PCWSTR name);
    ~Module();

    PCWSTR GetName() const {return m_name;}

    void AddTest(_In_ Test *test);
    void AddRun(_In_ PCWSTR command);
    void Execute();

    void SetLibrary(_In_ PAL_Library *lib);
private:
    size_t Find(_In_ PCWSTR test) const;
    void RunTests(size_t start);

    PCWSTR m_name;
    PAL_Library *m_library;      //Freed after the test run.
    vector<Test *> m_tests; //List of all tests in this module.
    vector<PWSTR> m_runs;  //List of selected tests to run.
};

NITS_EXPORT bool NitsParseArgument(_In_z_ char *str);
NITS_EXPORT int NitsExecuteRun();

class Run
{
public:
    static Run &GetInstance();
    static bool IsValid();

    Run();
    ~Run();

    Module &GetCurrentModule() const {return *m_loadingModule;}
    Test &GetCurrentTest() const {return *m_currentTest;}

    PCWSTR GetParam(_In_ PCWSTR name) const;
    PCWSTR GetDebugger() const {return m_debugger;}
    bool GetEnabled() const {return m_enabled;}
    bool GetFlaky() const {return m_flaky;}
    NitsFaultMode GetFaultMode() const {return m_simMode;}
    vector<int> const &GetFaultBreakpointList() const {return *m_bpFault;}
    Configuration const &GetConfiguration() const {return m_config;}

    bool IsParent() const {return !m_child;}
    bool IsChild() const {return m_child || !m_isolation;}
    bool IsSiteExcluded(int id) const;

    int  Execute();
    void ExecuteTest(_In_ Test *test, _In_opt_ PCWSTR choices);

    void SetDefaultOptions();
    void SetEnabled(bool enabled) {m_enabled = enabled;}
    void SetFlaky(bool flaky) {m_flaky = flaky;}
    void SetFaultMode(NitsFaultMode mode) {m_simMode = mode;}
    void SetExclusions(_In_reads_opt_(count) int sites[], int count);
    void ReportResult();
    bool ParseArgument(_In_ PWSTR arg);
private:
    static Run *s_run;

    void AddModule(_In_ PCWSTR name);
    bool AddParam(_In_ PCWSTR name, _In_ PCWSTR value);
    bool AddRun(_In_ PCWSTR name, _In_ PCWSTR test);
    void PipeThread();
    unsigned ReportStatistics();

    vector<Module *> &Modules() {return *m_modules;}
    vector<Param *> &Params() {return *m_params;}
    vector<Module *> const &Modules() const {return *m_modules;}
    vector<Param *> const &Params() const {return *m_params;}
#ifdef _MSC_VER
    static DWORD WINAPI _PipeThread(_In_ void *);
#else
    static void *_PipeThread(_In_ void *);
#endif

    Test *m_currentTest;
    Module *m_loadingModule;
    vector<Module *> *m_modules;
    vector<Param *> *m_params;

    //New options from command line.
    vector<int> *m_bpFault;     //List of fault iterations to break on.
    Configuration m_config;
    bool m_pause;               //Wait for input after testing?
    bool m_reset;               //Ignore existing test runs.
    PCWSTR m_debugger;          //Debug command line.
    PCWSTR m_binaryFilter;      //Semicolon-separated list; processes must contain one or more.
    PCWSTR m_binaryTarget;      //Semicolon-separated list of binaries to trap.

    //Old options from command line.
    bool m_isolation;           //Run tests in child process?
    bool m_child;               //Run as child process?
    bool m_fault;               //Run automatic fault sim (affects V1 tests only!).
    PCWSTR m_wtt;               //WTT log file.

    bool m_finished;            //Stops pipe thread.

#ifdef _MSC_VER
    CWTTLogger *m_logger;
    long m_loggerHandle;
#endif

    int *m_statistics;
    int m_faultIterations;      //Cumulative fault injection iteration total.

    //TODO: Deprecate all these variables.
    //Run-time options set by Test object.
    bool m_enabled;             //If false, skips the test body.
    bool m_flaky;               //If true, skips the test body only if nits was called with parameter "-skipflaky".
    NitsFaultMode m_simMode;    //Default to Auto/Manual fault sim.

    int *m_exclusions;          //Sites waived during auto fault sim.
    int m_exclusionsCount;
};

} //namespace TestSystem;
