//*****************************************************************************
//  Copyright (C) 2007 Microsoft Corporation
//  All rights reserved.
//*****************************************************************************

#include "Run.h"
#include <iostream>
#ifdef _MSC_VER
    #include <Switch.tmh>
#endif

using namespace std;
using namespace TestSystem;

#define IsSetupFixture(fixtureType) ((fixtureType == SetupFixture) || (fixtureType == SplitFixture))

NITS_EXPORT void NITS_CALL NitsSwitchCreate(
    _Inout_ Switch *test,
    _In_ PCWSTR name,
    _In_ void (NITS_CALL *setup)(Switch &),
    bool group)
{
    test->m_name = name;
    test->m_setup = setup;
    test->m_cleanup = NULL;
    test->m_choice = group ? Switch::AllChildren : 0;
    test->m_index = 0;
    test->m_refs = 0;
    test->m_children = new ChildList;
    test->m_registration = NULL;
    test->m_prevLayerContinuation = NULL;
    test->m_sameLayerContinuation = NULL;
    test->m_prevSwitchOnSelectedTestStack = NULL;
    test->m_rootSwitch = NULL;
    test->m_dryRunToFindChildren = 0;
    test->m_someoneCalledOmit = 0;
    test->m_omittedInCurrentTestChoice = 0;
    test->m_switchState = TestSystem::Switch::NotRunYet;
    
    if (test->m_children == NULL)
    {
        FatalError();
    }
}

NITS_EXPORT void NITS_CALL NitsSwitchDelete(
    _Inout_ Switch *test)
{
    struct RegistrationInfo *r = (struct RegistrationInfo *) (test->m_registration);
    int count = 0;
    delete test->m_children;
    if(r)
    {
        if(r->arrayOfInitializers)
        {
            while(count < r->numberOfParents)
            {
                free(r->arrayOfInitializers[count]);
                count++;
            }
            free(r->arrayOfInitializers);
        }

        if(r->parentArray)
        {
            free(r->parentArray);
        }
        
        free(r);
    }
}

NITS_EXPORT void NITS_CALL NitsSetTeardownSwitch(_Inout_ Switch *test, Switch *teardownSwitch, int isCleanup)
{
    if(test->m_registration)
    {
        struct RegistrationInfo *r = (struct RegistrationInfo *) (test->m_registration);
        if(isCleanup)
        {
            r->cleanupSwitch = teardownSwitch;
        }
        else
        {
            r->crashSwitch = teardownSwitch;
        }
    }
}

NITS_EXPORT void ** NITS_CALL NitsGetInitArray(_Inout_ Switch *test)
{
    if(!test->IsNewInterfaceTest())
        return NULL;
    
    struct RegistrationInfo *r = (struct RegistrationInfo *) (test->m_registration);

    return r->arrayOfInitializers;
}

NITS_EXPORT int NITS_CALL NitsIsSplitFixture(_In_ Switch *test)
{
    if(!test->IsNewInterfaceTest())
        return 0;
    
    struct RegistrationInfo *r = (struct RegistrationInfo *) (test->m_registration);    

    return (r->fixtureType == SplitFixture);
}

bool Switch::IsChildSelected(int index) const
{
    return IsGroup() || index == m_choice;
}

//Registers and runs child switches from setup().
//Caller must not change the child sequence when repeating its own variations.
NITS_EXPORT void NITS_CALL NitsSwitchChild(
    _Inout_ Switch *test,
    _In_ Switch &child)
{
    vector<Switch *> &list = *test->m_children;

    if (test->m_index == list.size())
    {
        //Register a new child.
        list.push_back(&child);
    }
    else if (list[test->m_index] != &child)
    {
        //Caller previously registered children in a different order!
        FatalError();
    }

    if (test->IsChildSelected(test->m_index++))
    {
        child.RunSetup();
    }
}

NITS_EXPORT bool NITS_CALL NitsSwitchGetEnabled(
    _Inout_ Switch const *test)
{
    return Run::GetInstance().GetEnabled();
}

NITS_EXPORT void NITS_CALL NitsSwitchSetFaultMode(
    _Inout_ Switch *test,
    NitsFaultMode mode)
{
    Run::GetInstance().SetFaultMode(mode);
}

NITS_EXPORT void NITS_CALL NitsSwitchSetEnabled(
    _Inout_ Switch *test,
    bool enabled)
{
    Run::GetInstance().SetEnabled(enabled);
}

NITS_EXPORT void NITS_CALL NitsSwitchSetFlaky(
    _Inout_ Switch *test,
    bool flaky)
{
    Run::GetInstance().SetFlaky(flaky);
}

NITS_EXPORT void NITS_CALL NitsSwitchFaultExclude(
    _Inout_ Switch *test,
    _In_reads_opt_(count) int sites[],
    int count)
{
    Run::GetInstance().SetExclusions(sites, count);
}

NITS_EXPORT void NITS_CALL NitsSwitchFaultOff(
    _Inout_ Switch *test)
{
    GetGlobals().SetFault(CallSite_NONE, 0, S_OK, L"");
}

NITS_EXPORT void NITS_CALL NitsSwitchFaultHresult(
    _Inout_ Switch *test,
    int site,
    HRESULT error,
    int attempt)
{
    GetGlobals().SetFault(site, attempt, error, L"");
}

NITS_EXPORT void NITS_CALL NitsSwitchFaultError(
    _Inout_ Switch *test,
    int site,
    DWORD error,
    int attempt)
{
    GetGlobals().SetFault(site, attempt, HRESULT_FROM_WIN32(error), L"");
}

NITS_EXPORT void NITS_CALL NitsSwitchFaultWait(
    _Inout_ Switch *test,
    int site,
    _In_ Event const &event,
    int attempt)
{
    GetGlobals().SetFault(site, attempt, S_OK, event);
}

void Switch::RunSetup()
{
    if (++m_refs > 1)
    {
        if (IsChoice())
        {
            //Sanity check: choices must NOT have multiple parents.
            //This causes confusion when choosing children.
            throw Exception();
        }

        return;
    }

    //The setup call registers and calls children through Child().
    //Child call order is determined by setup.
    m_setup(*this);
}

void Switch::RunCleanup()
{
    if (m_refs <= 0)
    {
        throw Exception();
    }

    if (--m_refs > 0)
    {
        return;
    }
    
    //Shut down the parent first.
    //Then shut down children that called setup, in reverse order.
    if (m_cleanup)
    {
        m_cleanup(*this);
    }

    while (m_index > 0)
    {
        vector<Switch *> &list = *m_children;
        Switch &child = *list[--m_index];
        if (IsChildSelected(m_index))
        {
            child.RunCleanup();
        }
    }
}

Switch *Switch::GetFirstSelectedChild()
{
    vector<Switch *> &list = *m_children;
    int index = 0;
    int sizeOfChildrenList = list.size();
    
    while (index < sizeOfChildrenList)
    {
        if (IsChildSelected(index))
        {
            return list[index];
        }
        index++;
    }

    return NULL;
}

void Switch::Reset()
{
    if (IsChoice())
    {
        m_choice = 0;
    }

    if (m_index != 0 || m_refs > 0)
    {
        throw Exception();
    }

    //TODO: Is this really necessary?
    vector<Switch *> &list = *m_children;
    for (DWORD i = 0; i < list.size(); i++)
    {
        list[i]->Reset();
    }

    m_prevLayerContinuation = NULL;
    m_sameLayerContinuation = NULL;
    m_prevSwitchOnSelectedTestStack = NULL;
    m_rootSwitch = NULL;
    m_switchState = TestSystem::Switch::NotRunYet;

    list.clear(); //Allows a later test to declare different children.
}

bool Switch::Next()
{
    vector<Switch *> &list = *m_children;
    if (IsGroup())
    {
        //Lexicographic search.
        //Start with the lowest-order "digit" and increment.
        //If that is done, reset and move to the next-order "digit".
        for (int i = list.size() - 1; i >= 0; i--)
        {
            Switch &child = *list[i];
            if (child.Next())
            {
                return true;
            }
            
            child.Reset();
        }

        return false;
    }
    else
    {
        if (list.size() == 0)
        {
            //Choice nodes with no children run like a group node.
            return false;
        }

        //Iterative search.
        //Each child contains at least one position.
        Switch &child = *list[m_choice];
        if (child.Next())
        {
            return true;
        }

        child.Reset();
        return ((unsigned long)(++m_choice)) < list.size();
    }
}

void Switch::PrintChoices(Buffer &data)
{
    wostringstream &buf = data;
    vector<Switch *> &list = *m_children;
    if (IsChoice())
    {
        if (list.size() > 0)
        {
            buf << L"/" << list[m_choice]->GetName();
            list[m_choice]->PrintChoices(data);
        }
    }
    else for (DWORD i = 0; i < list.size(); i++)
    {
        list[i]->PrintChoices(data);
    }
}

bool Switch::SelectChoices(_Inout_ PWSTR &choices)
{
    vector<Switch *> &list = *m_children;

    //NOTE: Have to run setup at least once to discover children.
    if (list.size() == 0)
    {
        RunSetup();
        RunCleanup();
    }

    //Recursively traverse the choices and groups.
    if (IsChoice())
    {
        PWSTR item = choices;
        PWSTR slash = wcschr(choices, L'/');
        if (slash)
        {
            //Advance to next choice.
            *slash = L'\0';
            choices = slash + 1;
        }
        else
        {
            //No more choices. Just advance to the end.
            choices += wcslen(choices);
        }

        for (DWORD i = 0; i < list.size(); i++)
        {
            if (Equals(list[i]->GetName(), item))
            {
                m_choice = i;
                return list[i]->SelectChoices(choices);
            }
        }

        //Choice was not found.
        return false;
    }
    else
    {
        for (DWORD i = 0; i < list.size(); i++)
        {
            if (!list[i]->SelectChoices(choices))
            {
                return false;
            }
        }

        return true;
    }
}

NITS_EXPORT void NITS_CALL NitsTestCreate(
    _Inout_ Test *test,
    _In_ PCWSTR name,
    _In_ void (NITS_CALL *setup)(Switch &),
    _In_ void (NITS_CALL *body)(Switch &))
{
    test->m_body = body;
    if (!Run::IsValid())
    {
        return;
    }

    Module &module = Run::GetInstance().GetCurrentModule();
    if (&module == NULL)
    {
        return;
    }

    test->m_module = module.GetName();
    module.AddTest(test);

    test->m_timeout = test->s_nextTimeout;
    test->m_isolation = test->s_nextIsolation;

    test->s_nextTimeout = 60;
    test->s_nextIsolation = false;
    test->m_startTicks = 0;
    test->m_testRunState = TestSystem::Test::BodyYetToRun;
}

void Test::PrintName(Buffer &data, bool choices)
{
    wostringstream &buf = data;
    buf << GetModule() << L":" << GetName();
    if (choices)
    {
        PrintChoices(data);
    }
}

void Test::Execute(_In_opt_ PCWSTR choices)
{
    Reset();
    Run &run = Run::GetInstance();
    run.SetDefaultOptions();

    if (choices == NULL)
    {
        do {
            RunVariation();
        }
        while (Next());
    }
    else
    {
        PWSTR temp = Copy(choices);
        PWSTR temp2 = temp;
        SetDryRun(1);
        bool success = SelectChoices(temp2) && temp2[0] == L'\0';
        SetDryRun(0);
        delete [] temp;

        if (!success)
        {
            wcerr << L"ERROR: Invalid test variation '" << choices << L"'\n";
            return;
        }

        RunVariation();
    }
}

void Test::ContinueProcessingAfterBody()
{
    if(m_testRunState == PartAfterBodyRan)
        return;

    m_testRunState = PartAfterBodyRan;


    Globals &globals = GetGlobals();
    Run &run = Run::GetInstance();

    Fault &fault = globals.GetAutoFault();
    bool failTest = false;

    if (globals.GetResult() == Error)
    {
        //Asserts attempted to fail fault sim.
        //During the regular body this is an ordinary failure.
        globals.SetResult(Failed);
    }
    else if (globals.GetResult() == Skipped && Run::GetInstance().GetEnabled())
    {
        if(!NitsAreWeInOmitMode())
        {
            //No asserts were attempted and the test wasn't disabled.
            globals.SetResult(Error);
        }
        else
        {
            globals.SetResult(Omitted);
        }
    }

    int tempSwitchState = m_switchState;
    
    bool runFaultInjection = Run::GetInstance().GetFaultMode() == Automatic &&
        globals.GetConfiguration().mode >= NitsTestIterativeFault;

    Configuration const &config = Run::GetInstance().GetConfiguration();

    bool filteringDetected = false;
    int limit = 0;

    if (!runFaultInjection || 
        globals.GetResult() != Passed)
    {
        goto NoFaultSim;
    }

    enum
    {
        MainThreadOnly,
        IncreasingAttemptOnly,
        Delayed
    } state;

    state = MainThreadOnly;
    globals.m_simAuto.m_minimumAttemptDifferentThread = INT_MAX;

    //Automatic fault simulation loop.
    for (int i = 1; true; i++)
    {
        globals.ResetIgnoredErrorReporting();
        if (state == Delayed)
        {
            //When filtering is disabled, threads from previous iterations
            //  might interfere with us. Wait some time for them to go away.
            PAL_Sleep(10);
        }

        //Do NOT reset globals here, or information from setup will be erased.
        //Set automatic fault sim parameters.
        bool faultBP = false;

        if (config.breakFault)
        {
            vector<int> const &list = Run::GetInstance().GetFaultBreakpointList();
            for (unsigned j = 0; j < list.size(); j++)
            {
                if (list[j] == i)
                {
                    faultBP = true;
                    break;
                }
            }
        }
        globals.SetResult(Passed);
        fault.Reset(CallSite_ALL, i, faultBP);

        wostringstream buf;
        buf << L"Starting fault simulation iteration #" << i << L"...";
        NitsTrace(buf.str().c_str());
#ifdef _MSC_VER
        DoTraceMessage(TestLifetime, L"Starting fault simulation iteration #%d...", i);
#endif
        globals.GetStats().faultIterations++;

        // making switch state NotRunYet so that running the body does not skip the body
        m_switchState = NotRunYet;

        m_body(*this);

        // putting the state back to what it was before we started fault sim so that further code path continues from the place where fault sim started
        m_switchState = tempSwitchState;
            
        if (globals.m_simAuto.m_filtered)
        {
            filteringDetected = true;
        }

        if (state == MainThreadOnly)
        {
            limit = max(limit, globals.m_simAuto.m_firstAttemptDifferentThread);
            globals.m_simAuto.m_firstAttemptDifferentThread = 0;
        }
        else if (state == IncreasingAttemptOnly)
        {
            limit = max(limit, globals.m_simAuto.m_faultedAttempt + 1);
            globals.m_simAuto.m_minimumAttemptDifferentThread = limit;
            globals.m_simAuto.m_firstAttemptDifferentThread = 0;
        }

        if (globals.GetResult() == Failed)
        {
            if (fault.DidFault())
            {
                //Expected behavior.
                //Proceed to the next iteration.
                continue;
            }

            //The test hit asserts before the fault was simulated.
            //Stop here and report failure.
            failTest = true;
            NitsTraceEx(L"UNEXPECTED FAULT SIM FAILURE", NitsHere(), NitsManual);
#ifdef _MSC_VER
            DoTraceMessage(TestLifetime, L"UNEXPECTED FAULT SIM FAILURE");
#endif
        }
        else if (globals.GetResult() == Error)
        {
            //Asserts attempted to fail fault sim.
            //Don't abort the test.
            failTest = true;
            continue;
        }
        else if (globals.GetResult() != Passed)
        {
            throw Exception();
        }

        //The test body passed.
        //This may or may not be expected.
        if (!fault.DidFault())
        {
            if (filteringDetected)
            {
                if (state == MainThreadOnly)
                {
                    //Rerun with all threads, using the calculated limit.
                    globals.m_simAuto.m_minimumAttemptDifferentThread = limit;
                    state = IncreasingAttemptOnly;
                }
                else if (state == IncreasingAttemptOnly)
                {
                    //Need to rerun with delays and without filtering.
                    globals.m_simAuto.m_minimumAttemptDifferentThread = 0;
                    state = Delayed;
                }
                filteringDetected = false;
                continue;
            }

            //Loop complete.
            fault.Reset(CallSite_NONE, 0, false);
            globals.SetResult(failTest ? Failed : Faulted);
#ifdef _MSC_VER
            DoTraceMessage(TestLifetime, L"Fault simulation loop completed");
#endif
            goto NoFaultSim;
        }

        if (Run::GetInstance().IsSiteExcluded(fault.m_hit) ||
            !globals.ShouldReportIgnoredErrors())
        {
            //The test expected faults not to propagate properly here.
            continue;
        }

        //This iteration should not have succeeded.
        //Report the problem and continue running the test.
        failTest = true;
        if (GetGlobals().GetConfiguration().traces >= NitsTraceAsserts)
        {
            NitsTraceEx(L"UNEXPECTED FAULT SIM SUCCESS",
                    fault.GetFaultedSite(), NitsManual);
        }
#ifdef _MSC_VER
        DoTraceMessage(TestLifetime, L"UNEXPECTED FAULT SIM SUCCESS");
        if (config.breakAssert)
        {
            DebugBreak();
        }
#endif
    }

NoFaultSim:

    Result temp = globals.GetResult();
    globals.SetResult(Skipped);

    wostringstream testName;

    Buffer wrappedTestName(testName);
    PrintName(wrappedTestName, true);

#ifdef _MSC_VER
    DoTraceMessage(TestLifetime, "Running test cleanup: %S", testName.str().c_str());
#endif

    RunCleanup();

    if(NitsAreWeInOmitMode())
    {
#ifdef _MSC_VER
        DoTraceMessage(TestLifetime, "Test choice ommited: %S", testName.str().c_str());
#endif
    }
    else
    {
#ifdef _MSC_VER
        DoTraceMessage(TestLifetime, "Test completed: %S", testName.str().c_str());
#endif
    }
    
#ifdef _MSC_VER
    //Reset all shared memory helpers except the NITS shared memory.
    //This prevents badly-behaved tests from interfering with future tests.
    System::GetInstance().ResetClientMappings();
#endif

    if (globals.GetResult() == Failed)
    {
        globals.SetResult(Error);
    }
    else
    {
        globals.SetResult(temp);
    }

    uint64 stopTicks = PAL_GetCurrentTime();
    double seconds = float(stopTicks - m_startTicks) / 1000.0;
    globals.SetRunTime(seconds);

    run.ReportResult();    
}

void Test::ContinueProcessingAfterSetup()
{
    if(m_testRunState == BodyRan)
        return;

    m_testRunState = BodyRan;
    
    Globals &globals = GetGlobals();
    Run &run = Run::GetInstance();

    wostringstream testName;

    Buffer wrappedTestName(testName);
    PrintName(wrappedTestName, true);
#ifdef _MSC_VER
    DoTraceMessage(TestLifetime, "Running test setup: %S", testName.str().c_str());
#endif

    if (globals.GetConfiguration().traces >= NitsTraceAllTests)
    {
        wostringstream buf;
        buf << L"\t          " << testName.str().c_str() << L"...\n";

        globals.PostPipe(buf.str().c_str());
    }

    if (globals.GetResult() == Failed)
    {
        globals.SetResult(Error);
    }
    else if (globals.GetResult() == Passed)
    {
        globals.SetResult(Skipped);
    }

    BOOL testEnabled = TRUE;
    if (!run.GetEnabled())
    {
        testEnabled = FALSE;
    }
    if (globals.GetConfiguration().skipFlakyTests)
    {
        if (run.GetFlaky())
        {
            testEnabled = FALSE;
        }
    }

    if (testEnabled &&
            globals.GetResult() == Skipped)
    {
#ifdef _MSC_VER
        DoTraceMessage(TestLifetime, "Running test: %S", testName.str().c_str());
#endif
        RunBody();
        
        if(m_testRunState == BodyRan)
        {
            if(!IsNewInterfaceTest())
            {
                ContinueProcessingAfterBody();
            }
            else
            {
                // we do not want test writers to use "return;" to exit the test fixtures, we want them to use "NitsReturn;"
                // if they use "return;" then we hit this code path where the Setup is not followed by a continuation function call leading to the body and so on.
                // so if we hit this code path when running new interface tests, we should error out
                throw Exception();
            }            
        }
    }
    else
    {
#ifdef _MSC_VER
        DoTraceMessage(TestLifetime, "Skipping test body: %S", testName.str().c_str());
#endif
        ContinueProcessingAfterBody();
    }
}


void Test::RunVariation()
{
    Run &run = Run::GetInstance();

    //Copy settings from command line options.
    run.SetDefaultOptions();

    m_startTicks = PAL_GetCurrentTime();

    m_testRunState = BodyYetToRun;
    m_prevLayerContinuation = NULL;
    m_sameLayerContinuation = NULL;
    m_prevSwitchOnSelectedTestStack = NULL;
    m_someoneCalledOmit = 0;
    m_switchState = TestSystem::Switch::NotRunYet;
    
    //Status is 'Skipped'.
    RunSetup();

    if(m_testRunState == BodyYetToRun)
    {    
        vector<Switch *> &list = *(m_children);
        // if it is not new interface test or there were no children registered then we will continue from here since setup would not have ran continuation
        if(!IsNewInterfaceTest() || (list.size() == 0))
        {
            ContinueProcessingAfterSetup();
        }
        else
        {
            // we do not want test writers to use "return;" to exit the test fixtures, we want them to use "NitsReturn;"
            // if they use "return;" then we hit this code path where the Setup is not followed by a continuation function call leading to the body and so on.
            // so if we hit this code path when running new interface tests, we should error out
            throw Exception();
        }
    }
}

void NO_INLINE Test::RunBody()
{
    Globals &globals = GetGlobals();
    
    //N.B. When nits.exe is running tests in-proc, we need to bootstrap
    //  the debugger before the test. Otherwise, the test could call
    //  LoadLibrary, and CheckDebugger() could get called inside the
    //  loader lock of our process, which can prevent us from actually
    //  creating the debugger process indefinitely.
    globals.CheckDebugger();

    globals.m_simAuto.m_firstAttemptDifferentThread = 0;
    globals.m_simAuto.m_minimumAttemptDifferentThread = 0;
    globals.m_simAuto.m_mainThread = PAL_GetCurrentThreadId();
    globals.m_simAuto.m_faultedAttempt = 0;

    //Run the regular test body.
    m_body(*this);
}

int Test::s_nextTimeout = 60;
bool Test::s_nextIsolation = false;

NITS_EXPORT void NITS_CALL NitsTestSetIsolation(_In_ Test &test)
{
    Test::s_nextIsolation = true;
}

NITS_EXPORT void NITS_CALL NitsTestSetTimeout(_In_ Test &test, int seconds)
{
    Test::s_nextTimeout = seconds;
}

NITS_EXPORT void NITS_CALL NitsSetRegistrationInfo(struct RegistrationInfo *r, 
                        int selfContextSize, 
                        int selfStructSize,                         
                        int numberOfParents,
                        Switch **parentArray, 
                        void *(*creationFunc)(), 
                        void (*initFunc)(void *, void *), 
                        void (*bodyFunc)(void *),
                        void (*cleanupFunc)(void *),
                        enum TestFixtureType fixtureType, 
                        void **arrayOfInitializers)
{
    r->selfContext = 0;
    r->selfStruct = 0;    
    r->countOfLinks = 0;
    r->selfContextSize = selfContextSize;
    r->selfStructSize = selfStructSize;    
    r->numberOfParents = numberOfParents;
    r->parentArray = parentArray;
    r->creationFunc = creationFunc;
    r->initFunc = initFunc;
    r->bodyFunc = bodyFunc;
    r->cleanupFunc = cleanupFunc;
    r->fixtureType = fixtureType;
    r->arrayOfInitializers = arrayOfInitializers;        
    r->cleanupSwitch = 0;
    r->crashSwitch = 0;
}

NITS_EXPORT void NITS_CALL NitsCreateStruct(TestSystem::Switch &myFixture, void **selfContext, void **selfStruct)
{
    struct RegistrationInfo *myFixtureRegistration = (struct RegistrationInfo *) (myFixture.m_registration);

    // this should not happen
    if(!selfContext || !selfStruct)
        return;

    // this is for virtual inheritance kind of semantics; if there is already a structure allocated through one route in inheritance, we will not allocate it again but use the same one
    if(myFixtureRegistration->countOfLinks > 0)
    {
        *selfContext = myFixtureRegistration->selfContext;
        *selfStruct = myFixtureRegistration->selfStruct;
    }
    else
    {
        int isSplitFixture = (myFixtureRegistration->fixtureType == SplitFixture);
        // for splitfixtures, the context comes from their parents; they do not have a context; 
        // pointer adjustment is done during the call of the splitfixture by walking back the call chain
        *selfContext = malloc(myFixtureRegistration->selfContextSize);
        *selfStruct = (isSplitFixture ? 0 : malloc(myFixtureRegistration->selfStructSize));
        myFixtureRegistration->selfContext = *selfContext;
        myFixtureRegistration->selfStruct = *selfStruct;
    }
    (myFixtureRegistration->countOfLinks)++;
}

NITS_EXPORT void NITS_CALL NitsDestroyStruct(TestSystem::Switch &myFixture)
{
    struct RegistrationInfo *myFixtureRegistration = (struct RegistrationInfo *) (myFixture.m_registration);
    (myFixtureRegistration->countOfLinks)--;

    if(myFixtureRegistration->countOfLinks == 0)
    {
        free(myFixtureRegistration->selfContext);

        if(myFixtureRegistration->fixtureType != SplitFixture)
        {
            free(myFixtureRegistration->selfStruct);
        }
        myFixtureRegistration->selfContext = 0;
        myFixtureRegistration->selfStruct = 0;
    }
}

NITS_EXPORT void * NITS_CALL NitsGetSelfStruct(TestSystem::Switch &myFixture)
{
    struct RegistrationInfo *myFixtureRegistration = (struct RegistrationInfo *) (myFixture.m_registration);

    return myFixtureRegistration->selfStruct;
}

NITS_EXPORT void NITS_CALL NitsBody_NewInterfaceTest(TestSystem::Switch &test)
{
    struct RegistrationInfo *testRegistration = (struct RegistrationInfo *) (test.m_registration);

    if(test.m_switchState == TestSystem::Switch::NotRunYet)
    {
        test.m_switchState = TestSystem::Switch::RunOnce;
        testRegistration->bodyFunc(testRegistration->selfContext);
    }
    else
    {
        NitsRunContinuation(test);
    }
}

NITS_EXPORT void NITS_CALL NitsSetup_NewInterfaceTest(TestSystem::Switch &test)
{
    struct RegistrationInfo *testRegistration = (struct RegistrationInfo *) (test.m_registration);
    int parentCount = 0;
    Switch *currentParent = 0;
    // when doing dryRun, it is needed only to make engine aware of children; so skip all other steps.
    int isDryRun = Run::GetInstance().GetCurrentTest().IsDryRunToIdentifyChildren();// test.IsDryRunToIdentifyChildren();
    
    if((testRegistration->fixtureType == BodyFixture) ||
        IsSetupFixture(testRegistration->fixtureType))
    {    
        if(testRegistration->fixtureType == BodyFixture && !isDryRun)
        {
            if(testRegistration->selfContext == 0)
            {
                // this creationFunc will recurse through the entire test structure and create selfContext for each node following the virtual inheritance way of creating common node only once
                // TODO: implement multiple inheritance; will the common nodes need to register switches multiple times; who will create those switches?
                testRegistration->creationFunc();
            }
            // the initFunc is passed in the created context and a NULL for the initValues since the test itself does not provide any init values; the initFunc is supposed zero out its structure if init values are not provided.
            testRegistration->initFunc(testRegistration->selfContext, 0);

            test.m_sameLayerContinuation = 0;

            test.m_prevLayerContinuation = &test;
        }

        if(testRegistration->parentArray != 0)
        {
            while(parentCount < testRegistration->numberOfParents)
            {
                // register the child by calling Test.Child; this helps the test engine identify children/pick up the current chain of children to be run in this iteration
                currentParent = testRegistration->parentArray[parentCount];
                
                if(wcscmp(currentParent->m_name, L"SplitFixture") != 0)
                {
                    // in both test setup and body cases, you need continuation switch set to your currentParent; 
                    // but in SetupFixture case the continuation is the body of this current testRegistration whereas in 
                    // in BodyFixture case continuation is the ContinueProcessingAfterSetup
                    if(!isDryRun)
                    {
                        currentParent->m_sameLayerContinuation = 0;
                        
                        if(test.IsGroup() && (parentCount >= 1))
                        {
                            Switch *prevParent = testRegistration->parentArray[parentCount - 1];
                            prevParent->m_sameLayerContinuation = currentParent;
                        }

                        // this helps with multiple inheritance; if some other fixture already put the prevLayerContiuationLink into current parent, that is sufficient; the second link is not required
                        if(currentParent->m_prevLayerContinuation == 0)
                        {
                            currentParent->m_prevLayerContinuation = &test;
                        }
                    }
                    
                    test.Child(*currentParent);
                }
                
                parentCount++;
            }

            // once we found what is the root node in current iteration; invoke its body; that will keep calling continuation till we run the test body and cleanup by nesting them on the same stack
            if(!isDryRun && testRegistration->fixtureType == BodyFixture && (test.m_rootSwitch != 0))
            {
                (test.m_rootSwitch)->m_prevSwitchOnSelectedTestStack = 0;
                NitsBody_NewInterfaceTest(*(test.m_rootSwitch));
                test.m_rootSwitch = 0;                
            }
        }
        else
        {
            if(!isDryRun)
            {
                // when we reach the root of the tree; run the body of the fixture
                if(IsSetupFixture(testRegistration->fixtureType))
                {
                    Switch *continuationSwitch = test.m_prevLayerContinuation;
                    while(continuationSwitch)
                    {
                        if(continuationSwitch->m_prevLayerContinuation == 0 || (continuationSwitch->m_prevLayerContinuation == continuationSwitch))
                        {
                            break;
                        }
                        continuationSwitch = continuationSwitch->m_prevLayerContinuation;
                    }

                    if (continuationSwitch == NULL)
                    {
                        FatalError();
                        return;
                    }

                    // only first root node i.e. when m_rootSwitch = 0 does this assignment; others do not since the first one will continue to run others
                    if(continuationSwitch->m_rootSwitch == 0)
                    {
                        continuationSwitch->m_rootSwitch = &test;
                    }
                }
            }
        }
    }
}

NITS_EXPORT void NITS_CALL NitsCleanup_NewInterfaceTest(TestSystem::Switch &test)
{
    struct RegistrationInfo *testRegistration = (struct RegistrationInfo *) (test.m_registration);

    if(Run::GetInstance().GetCurrentTest().IsDryRunToIdentifyChildren())
        return;
    
    struct RegistrationInfo *cleanupRegistration = (struct RegistrationInfo *) ((testRegistration->cleanupSwitch) ? testRegistration->cleanupSwitch->m_registration : 0);
    
    if(cleanupRegistration && !(test.m_omittedInCurrentTestChoice))
    {
        if(test.m_switchState == TestSystem::Switch::RunOnce)
        {
            test.m_switchState = TestSystem::Switch::CleanupRun;
            cleanupRegistration->bodyFunc(testRegistration->selfContext);
        }
    }

    test.m_switchState = TestSystem::Switch::NotRunYet;
    test.m_omittedInCurrentTestChoice = 0;
}

NITS_EXPORT void NITS_CALL NitsRunContinuation(TestSystem::Switch &test)
{
    // start psuedo code
    TestSystem::Switch *continuationSwitch = test.m_sameLayerContinuation;
    struct RegistrationInfo *continuationRegistration = 0;

    while(continuationSwitch)
    {
        continuationRegistration = (struct RegistrationInfo *) (continuationSwitch->m_registration);
        
        if(continuationRegistration->parentArray == 0)
        {
            continuationSwitch->m_prevSwitchOnSelectedTestStack = &test;
            continuationSwitch->m_omittedInCurrentTestChoice = Run::GetInstance().GetCurrentTest().m_someoneCalledOmit;
            if(continuationRegistration->fixtureType == SplitFixture)
            {
                continuationRegistration->selfStruct = NitsGetSelfStruct(test);
            }
            NitsBody_NewInterfaceTest(*continuationSwitch);
            // continuationRegistration->bodyFunc(continuationRegistration->selfContext);
            return;
        }
        else
        {
            Switch *tempSwitch = continuationSwitch;
            continuationSwitch = continuationSwitch->GetFirstSelectedChild();
            if(continuationSwitch && (continuationSwitch->m_prevLayerContinuation != tempSwitch))
            {
                // this helps with multiple inheritance; before going to the next parent; 
                // you set the prevLayerContinuation to yourself so that when it comes down, it does not go the current prevLayer instead comes back to you
                continuationSwitch->m_prevLayerContinuation = tempSwitch;
            }
        }
    }

    continuationSwitch = test.m_prevLayerContinuation;

    if(continuationSwitch)
    {
        continuationRegistration = (struct RegistrationInfo *) (continuationSwitch->m_registration);

        if(IsSetupFixture(continuationRegistration->fixtureType))
        {
            continuationSwitch->m_prevSwitchOnSelectedTestStack = &test;
            continuationSwitch->m_omittedInCurrentTestChoice = Run::GetInstance().GetCurrentTest().m_someoneCalledOmit;
            if(continuationRegistration->fixtureType == SplitFixture)
            {
                continuationRegistration->selfStruct = NitsGetSelfStruct(test);
            }            
            NitsBody_NewInterfaceTest(*continuationSwitch);
            // continuationRegistration->bodyFunc(continuationRegistration->selfContext);
        }
        else if(continuationRegistration->fixtureType == BodyFixture)
        {
            TestSystem::Test *continuationTest = (TestSystem::Test *)continuationSwitch;
            
            if(continuationTest->m_testRunState == TestSystem::Test::BodyYetToRun)
            {
                continuationSwitch->m_prevSwitchOnSelectedTestStack = &test;
                continuationSwitch->m_omittedInCurrentTestChoice = Run::GetInstance().GetCurrentTest().m_someoneCalledOmit;
                continuationTest->ContinueProcessingAfterSetup();
            }
            else if(continuationTest->m_testRunState == TestSystem::Test::BodyRan)
            {
                continuationTest->ContinueProcessingAfterBody();

                // the main thread which was running body and possibly faultsim will then call clean up on the test related context objects
                // this is separate from the cleanup functions explicitly defined by the user to clean up the stuff the user wants to.
                continuationRegistration->cleanupFunc(continuationRegistration->selfContext);
            }            
        }
    }
}

NITS_EXPORT int NITS_CALL NitsIsFixtureSelectedSoFar(TestSystem::Switch &currentSwitch, TestSystem::Switch &searchSwitch)
{
    Switch *currentNode = &currentSwitch;
    
    while(currentNode->m_prevSwitchOnSelectedTestStack)
    {
        if(currentNode->m_prevSwitchOnSelectedTestStack == &searchSwitch)
            return 1;

        currentNode = currentNode->m_prevSwitchOnSelectedTestStack;
    }

    return 0;
}

#define FIXTURE_STRING L"Fixture "
#define OMITTED_MESSAGE L" omitted the test choice"

NITS_EXPORT void NITS_CALL NitsOmit(TestSystem::Switch &currentSwitch)
{
    Run::GetInstance().GetCurrentTest().m_someoneCalledOmit = 1;
    int totalLen = wcslen(FIXTURE_STRING) + wcslen(currentSwitch.GetName()) + wcslen(OMITTED_MESSAGE) + 1;
    wchar_t *outString = (wchar_t *)(malloc(totalLen*sizeof(wchar_t)));
    if(outString)
    {
        PAL_CopyByLenW(outString, L"Fixture ", totalLen);
        PAL_ConcatByLenW(outString, currentSwitch.GetName(), totalLen);
        PAL_ConcatByLenW(outString, L" omitted the test choice", totalLen);
        TTRACE(outString);
        free(outString);
    }
    else
    {
        TTRACE(L"Some Fixture omitted the test choice");
    }
    GetGlobals().SetResult(Skipped);
}

NITS_EXPORT int NITS_CALL NitsAreWeInOmitMode()
{
    return Run::GetInstance().GetCurrentTest().m_someoneCalledOmit;
}

#ifdef _MSC_VER
    NITS_EXPORT struct EmptyStruct sEmptyStruct = {0};
#else
    __attribute__ ((visibility("default"))) struct EmptyStruct sEmptyStruct = {0};
#endif
