//*****************************************************************************
//  Copyright (C) 2007 Microsoft Corporation
//  All rights reserved.
//*****************************************************************************

#include "Globals.h"
#ifdef _MSC_VER
    #include <nitsdll.tmh>
#endif

namespace TestSystem
{

//This object sets up internal cross-process structures automatically.
//No other action is necessary beyond linking to this DLL.
System System::s_system;

NITS_EXPORT DWORD Initialize(_In_ PCWSTR name,
                 _In_reads_bytes_(bytes) void const *context,
                 _Outptr_result_bytebuffer_(bytes) void **shared,
                 long bytes)
{
    static long volatile s_spinlock = FALSE;
    while (PAL_InterlockedExchange(&s_spinlock, TRUE) == TRUE)
    {
        PAL_Sleep(10);
    }

    DWORD error = 0;

    Mapping &m = System::GetInstance().GetMapping(name);
    error = m.Initialize(context, bytes);
    *shared = m.m_data;

    PAL_InterlockedExchange(&s_spinlock, FALSE);
    return error;
}

NITS_EXPORT DWORD GetFaultWin32()
{
    return WIN32_FROM_HRESULT(GetGlobals().GetFaultError());
}

}//namespace TestSystem

//Prototype NITSv2 constructs.
struct Fixture;

struct Registration
{
    //Split this up into separate registrations for each type.
    Fixture *(*creator)(void (*func)(Fixture *, void *), void *context);
    void (*runner)(Fixture *, void (*next)(void *), void *context);
    int type;
    bool multiple;

    char *name;
    Registration *cleanup;
    Registration *crash;
};

struct Fixture
{
    //Can we use templates or something to get rid of this field?
    Registration *r;
    void Register(void (*func)(Fixture *, void *), void *context)
    {
        //Multiple-inheritance fixtures will appear in different locations.
        //Virtual-inheritance fixtures will appear multiple times in the same location.
        //func needs to discard virtual-inheritance duplicates!
        func(this, context);
    }
};

//Implement flags/settings as base classes!
//NOTE: multiple-inheritance of flags is OK if one overrides the other
//  but results in an error if ambiguous. Some flags automatically
//  accept the most restrictive of all seen combinations.
template<int arg> struct X1 : protected Fixture {int c;};
struct X2 : protected Fixture {int c;};

//Use separate macros for virtual/multiple setup.
//-Splitting fixtures must be virtual.
//-Templatized fixtures could be considered separate.
//-Name lookup needed only for tests, which aren't templatized.
//-NOTE: Need flag to fail multiple simultaneous templates of the same fixture.

//Can't do this with macros. Define Child_X1 and Child_X2 as separate classes instead and inherit from those!
#define _X1 virtual protected X1
#define _X2 protected X2

#define PARENTIZE2(a) _ ## a
#define PARENTIZE(a) PARENTIZE2(a)

//Use an extra layer of unique base classes to prevent multiple inheritance warnings.
#define CURRENT_CLASS Class_Foo
struct Parent_Foo : protected Fixture {};
struct Parent_Foo_1 : PARENTIZE(X1<5>) {};
struct Parent_Foo_2 : PARENTIZE(X2) {};
struct Class_Foo : protected Parent_Foo_1,
                   protected Parent_Foo_2,
                   protected Parent_Foo
{
    static Registration s_reg;
    static Fixture *Create(void (*func)(Fixture *, void *), void *context)
    {
        Class_Foo *p = new Class_Foo;
        p->Register(func, context);
        Parent_Foo *pp = p;
        return pp;
    }

    static void Execute(Fixture *f, void (*next)(void *), void *context)
    {
        //This reverses the this-pointer adjustments.
        Parent_Foo *pp = static_cast<Parent_Foo *>(f);
        Class_Foo *p = static_cast<Class_Foo *>(pp);
        p->Run(next, context);
    }

    void Register(void (*func)(Fixture *, void *), void *context)
    {
        //This handles the automatic this-pointer adjustments.
        Parent_Foo_1::Register(func, context);
        Parent_Foo_2::Register(func, context);
        Parent_Foo::Register(func, context);
    }

    void Run(void (*next)(void *), void *context) {
        //Area between SETUP and END_SETUP.
        x = X1::c;
        p = NULL;

        //End area between SETUP and END_SETUP
        next(context); //Capture current stack variables.
    }

    //Fields in END_SETUPx.
    int x;
    char *p;
};

//Set the data for the registration during SETUP, with macros.
//That way "Foo" doesn't need to be in END_SETUP.
//Can't do this: use templates instead to do it inside the class using a constructor.
Registration CURRENT_CLASS::s_reg = {&CURRENT_CLASS::Create, &CURRENT_CLASS::Execute};
#undef CURRENT_CLASS

struct Foo_Cleanup : protected Class_Foo
{
    //No new fields, so data format is the same and casting is OK.
    //Cleanup/crash fixtures do not have any additional parents!
};

struct Foo_Crash : protected Class_Foo
{
    //No new fields, so data format is the same and casting is OK.
    //Sometimes OK to use variables here but be careful!
    //When an exception happens inside Foo, Crash_Foo is run, and then
    //  the exception is rethrown. (Simpler than trying cleanup fixtures.)
    //When a process dies abruptly, the crash fixture is run for all
    //  fixtures known to be pending.
};

//Test fixtures: same as setup but no cleanup/crash children and no fields.
//Test fixtures cannot have templates or children but may have parents/flags.
 
//Splitting fixtures: normally use a common virtual base class.
//Otherwise, only one set of variables will be set at a time.
//Splitting fixtures call OmitParent to selectively omit parents.
//If all are omitted, this variation is skipped over.
//If OmitParent is not called, all parents are run.
//The parent set must be known up front for simplicity.

//NOTE: Want conditional fixtures!
//Example: fixture A has a split-dependency on {B, C}, and A has additional step D to run if B is selected.
//Something like: Conditional<B, D>
//
//AND: Allow Reference<B> to refer (virtually) to B's variables without instantiating it, and call B::IsSelected() as the condition.
