
#include "../base/Globals.h"

#ifdef _MSC_VER
    #include <psapi.h>
#endif

namespace TestSystem {

Fault::Fault() : m_lock(0)
{
    Reset(CallSite_NONE, 0, false);

    m_mainThread = 0;
    m_minimumAttemptDifferentThread = 0;
}

void Fault::Reset(int site, int iteration, bool breakOnFault)
{
    m_site = site;
    m_iteration = iteration;
    m_attempt = 0;

    m_break = breakOnFault;
    m_faulted = false;
    m_filtered = false;
    m_hit = CallSite_NONE;
    m_line = 0;
    m_file[0] = '\0';

    m_faultedAttempt = 0;
    m_firstAttemptDifferentThread = 0;
}

Globals::Globals()
    : m_version(SharedMemoryVersion),
      m_runLock(0),
      m_attachLock(0),
      m_pipeLock(FALSE),
      m_pipeChars(0)
{
    m_pipe[0] = '\0';
    m_debugger[0] = '\0';
    m_binaryFilter[0] = '\0';
    m_binaryTarget[0] = '\0';

    for (int i = 0; i < InjectorListSize; i++)
        m_injectors[i].process = 0;

    Reset();

    for (int i = 0; i < ResultCount; i++)
    {
        m_statistics[i] = 0;
    }
}

Globals::~Globals()
{
}

void Globals::Reset()
{
    m_result = Skipped;

    m_stopReportingIgnoredErrors = false;

    m_config.traces = NitsTraceAllTests;
    m_config.mode = NitsTestCompoundFault;
    m_config.breakFault = false;
    m_config.breakAssert = false;
    m_config.skipFlakyTests = false;

    m_simAuto.Reset(CallSite_NONE, 0, false);
    m_simManual.Reset(CallSite_NONE, 0, false);

    m_faultError = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    m_faultEvent[0] = L'\0';
    m_debugger[0] = L'\0';
    m_binaryFilter[0] = '\0';
    m_binaryTarget[0] = '\0';
}

} //namespace TestSystem

// I have ported this as much as I could; main remaining things are EnumProcessModules/GetModuleBaseName
// which I did not find equivalents; so I would prefer to leave this code in windows only
// also we have not finalized on the method we will be using to have product call into framework apis
// so I will defer this work to that point when we decide how linux product will call into framework
#ifdef _MSC_VER
static void PatchBinary(
    _In_z_ char *binary)
{
    PAL_Library *library;
    PAL_Library *framework;
    NitsFT *target;
    NitsFT *source;

    library = PAL_LoadLibraryA(binary);
    if (library == NULL)
    {
        return;
    }

    target = (NitsFT *)PAL_GetProcAddress(library, "NITS");
    if (target == NULL)
    {
        PAL_FreeLibrary(library);
        return;
    }

#ifdef _MSC_VER
    framework = PAL_LoadLibraryA("nitsdll.dll");
#else
    framework = PAL_LoadLibraryA("libnitsdll.so");
#endif
    if (framework == NULL)
    {
        return;
    }

    source = (NitsFT *)PAL_GetProcAddress(framework, "NITS");
    if (source == NULL)
    {
        PAL_FreeLibrary(library);
        return;
    }

    memcpy(target, source, sizeof(NitsFT));
    PAL_FreeLibrary(library);
}

static unsigned DoesBinaryMatch(
    _In_ char *list,
    _In_z_ char *name)
{
    for (;;)
    {
#pragma prefast(suppress: 26006)
        if (*list == '\0')
            return FALSE;

#pragma prefast(suppress: 26006)
        if (_stricmp(list, name) == 0)
            return TRUE;

        list += strlen(list) + 1;
    }
}


void ProcessPatches(TestSystem::Globals *globals)
{
    if (globals->m_runLock == 0)
        return;

    /* Sweep through the list of loaded modules. */
    HMODULE modules[200];
    DWORD size;
    unsigned count;
    unsigned i;
    unsigned isMatch = 0;
    char name[MAX_PATH];

    if (EnumProcessModules(GetCurrentProcess(), modules, sizeof(modules), &size) == FALSE)
    {
        return;
    }

    count = size / sizeof(HMODULE);

    if (globals->m_binaryFilter[0] == '\0')
    {
        /* There is no filter. Match target binaries in any process. */
        isMatch = 1;
    }
    else for (i = 0; i < count; i++)
    {
        GetModuleBaseNameA(GetCurrentProcess(), modules[i], name, MAX_PATH);
        if (DoesBinaryMatch(globals->m_binaryFilter, name))
        {
            /* Found one of the filter modules. */
            isMatch = 1;
            break;
        }
    }

    if (isMatch == 0)
        return;

    for (i = 0; i < count; i++)
    {
        GetModuleBaseNameA(GetCurrentProcess(), modules[i], name, MAX_PATH);
        if (DoesBinaryMatch(globals->m_binaryTarget, name))
        {
            /* Found a target module. Trap this! */
            PatchBinary(name);
        }
    }
}

static TestSystem::Globals g_tempGlobals;

static PAL_Semaphore *g_signalSemaphore;
static PAL_Semaphore *g_waitSemaphore;
static TestSystem::Globals *g_globals = NULL;
static TestSystem::InjectorTarget *g_injectorTarget = NULL;

unsigned long InjectorSetup()
{

    PAL_SharedMemory *mapping;
    void *start;
    unsigned bytes = sizeof(TestSystem::Globals) + sizeof(long);
    volatile long *status;
    char nameSignal[50] = "NitsInjectorIn_";
    char nameWait[50] = "NitsInjectorOut_";

    /* TODO: portable shared memory mapping. */
    mapping = PAL_SharedMemory_Open(L"Global\\NitsGlobalData", bytes);
    if (mapping == NULL)
    {
        return GetLastError();
    }

    start = PAL_SharedMemory_Map(mapping, bytes);
    if (start == NULL)
    {
        return GetLastError();
    }

    /* The shared memory is mapped. Make sure the contents are initialized. */
    status = (long *)start;
    g_globals = (TestSystem::Globals *)(status + 1);
    if (PAL_InterlockedCompareExchange(status, TestSystem::Loading, TestSystem::Unloaded) == TestSystem::Unloaded)
    {
        memcpy(g_globals, &g_tempGlobals, sizeof(TestSystem::Globals));
        *status = TestSystem::Loaded;
    }

    while (*status != TestSystem::Loaded)
    {
        PAL_Sleep(10);
    }

    /* The shared memory space is now initialized. */
    /* Set up the semaphores used to trigger patching. */

    if (g_globals->m_version != TestSystem::SharedMemoryVersion)
    {
        g_globals = NULL;
        return 0;
    }

    _itoa_s(PAL_GetCurrentProcessId(), nameSignal + 15, sizeof(nameSignal) - 15, 10);
    _itoa_s(PAL_GetCurrentProcessId(), nameWait + 16, sizeof(nameWait) - 16, 10);

    /* The semaphre signalled by the product to us. */
    g_signalSemaphore = PAL_CreateSemaphore(nameSignal);

    /* The product waits on this semaphre to continue execution. */
    g_waitSemaphore = PAL_CreateSemaphore(nameWait);

    /* Register the handles with the harness. */
    for (int i = 0; i < TestSystem::InjectorListSize; i++)
    {
        TestSystem::InjectorTarget *target = g_globals->m_injectors + i;

        if (target->process != 0)
            continue;

        if (InterlockedCompareExchange(&target->process, PAL_GetCurrentProcessId(), 0) != 0)
            continue;

        target->signalSemaphore = g_signalSemaphore->m_handle;
        target->waitSemaphore = g_waitSemaphore->m_handle;
        g_injectorTarget = target;
        break;
    }

    return 0;
}

void Unload()
{
#ifdef _MSC_VER
    HMODULE hModule;
    void *address = Unload;
    if (GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS |
                GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT, (LPCTSTR)address, &hModule))
    {
        FreeLibraryAndExitThread(hModule, 0);
    }

    ExitThread(0);
#endif
}

unsigned long WINAPI InjectorProc(_In_ void *param)
{
    if (g_globals == NULL)
    {
        Unload();
        return 0;
    }

    for(;;)
    {
        PAL_WaitForSemaphore(g_signalSemaphore, INFINITE);

        if (g_globals->m_unload ||
            g_globals->m_version != TestSystem::SharedMemoryVersion ||
            g_injectorTarget->process != PAL_GetCurrentProcessId())
        {
            PAL_SignalSemaphore(g_waitSemaphore);
            Unload();
            return 0;
        }
        ProcessPatches(g_globals);

        PAL_SignalSemaphore(g_waitSemaphore);
    }

    return 0;
}

void InjectorCleanup()
{
    if (g_injectorTarget == NULL)
        return;

    if (g_injectorTarget->process == PAL_GetCurrentProcessId())
    {
        g_injectorTarget->process = 0;
    }
}
#endif
