/*
 * Copyright (C) 2012 Microsoft Corporation
 * All rights reserved.
 */

#define HOOK_BUILD

#ifdef _MSC_VER
    #include <windows.h>
#endif
#include <stdlib.h>
#include <nits.h>

// most of the porting is done except a few like PeekMessage/_itoa_s etc
// rest of the porting will be done when we decide which method we will be using
// for product to link into the framework

void SignalInjector()
{
#ifdef _MSC_VER
    char nameSignal[50] = "NitsInjectorIn_";
    char nameWait[50] = "NitsInjectorOut_";
    HANDLE semaphore;
    MSG unused;

    /* Dummy call to user32.dll. This forces the stubbed DLL to depend on
     * user32, which in turn supports bootstrapping the injector. */
    PeekMessage(&unused, NULL, 0, 0, PM_NOREMOVE);

    _itoa_s(GetCurrentProcessId(), nameSignal + 15, sizeof(nameSignal) - 15, 10);
    _itoa_s(GetCurrentProcessId(), nameWait + 16, sizeof(nameWait) - 16, 10);

    semaphore = PAL_OpenSemaphore(nameSignal);
    if (semaphore == NULL)
    {
        /* The injector isn't there. */
        return;
    }

    PAL_SignalSemaphore(semaphore);
    PAL_DestructSemaphore(semaphore);

    semaphore = PAL_OpenSemaphore(nameWait);
    if (semaphore == NULL)
    {
        /* The injector isn't there. */
        return;
    }

    /* Use a short timeout to prevent undesired behavior. In the worst
     * case, the injector thread will be starved and the patching will
     * happen later. */
    PAL_WaitForSemaphore(semaphore, 100);
    PAL_DestructSemaphore(semaphore);
#endif
}

void CheckInjector()
{
#ifdef _MSC_VER
    /* The injector might not know we are here.
     * Open up the injector refresh semaphore for this process.
     * Then signal it once and wait for the NitsFT to be patched. */
    static unsigned s_signalSent = 0;

    if (s_signalSent == 0)
    {
        /* Okay to hit this path more than once sometimes. */
        s_signalSent = 1;
        SignalInjector();

        /* Re-enter through the function table. This should be patched! */
    }

    /* The function table should be unchanged. The caller will attempt to
     * replace itself with the simple stub, since the check is complete. */
#endif
}


#define CompareSwapFunction(dest, src, cmp) \
    PAL_InterlockedCompareExchangePointer((void **)&dest, (void *)src, (void *)cmp)

static NitsResult NITS_CALL ShouldFault(
    NitsCallSite line,
    NitsFaultMode mode)
{
    return NitsFalse;
}

static NitsResult NITS_CALL ShouldFault_Checked(
    NitsCallSite line,
    NitsFaultMode mode)
{
    CheckInjector();
    CompareSwapFunction(NITS.ShouldFault, ShouldFault, ShouldFault_Checked);
    return NITS.ShouldFault(line, mode);
}

static void NITS_CALL BeginResource(
    _In_ void *resource,
         int type,
         NitsResourceScope scope,
         NitsCallSite line)
{
}

static void NITS_CALL BeginResource_Checked(
    _In_ void *resource,
         int type,
         NitsResourceScope scope,
         NitsCallSite line)
{
    CheckInjector();
    CompareSwapFunction(NITS.BeginResource, BeginResource, BeginResource_Checked);
    NITS.BeginResource(resource, type, scope, line);
}

static void NITS_CALL ChangeResource(
    _In_ void *resource,
         int type,
         NitsResourceScope scope,
         NitsCallSite line)
{
}

static void NITS_CALL EndResource(
    _In_ void *resource,
         int type,
         NitsCallSite line)
{
}

static void NITS_CALL TraceA(
    _In_z_ const char *text,
           NitsCallSite line,
           NitsFaultMode mode)
{
}

static void NITS_CALL TraceW(
    _In_z_ const wchar_t *text,
           NitsCallSite line,
           NitsFaultMode mode)
{
}

static NitsResult NITS_CALL AssertA(
           int test,
    _In_z_ const char *text,
    _In_z_ const char *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    return NitsTrue;
}

static NitsResult NITS_CALL AssertA_Checked(
           int test,
    _In_z_ const char *text,
    _In_z_ const char *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    CheckInjector();
    CompareSwapFunction(NITS.AssertA, AssertA, AssertA_Checked);
    return NITS.AssertA(test, text, description, line, mode);
}

static NitsResult NITS_CALL AssertW(
           int test,
    _In_z_ const char *text,
    _In_z_ const wchar_t *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    return NitsTrue;
}

static NitsResult NITS_CALL AssertW_Checked(
           int test,
    _In_z_ const char *text,
    _In_z_ const wchar_t *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    CheckInjector();
    CompareSwapFunction(NITS.AssertW, AssertW, AssertW_Checked);
    return NITS.AssertW(test, text, description, line, mode);
}

static NitsResult NITS_CALL CompareA(
           int lhs,
           int rhs,
    _In_z_ const char *lhsText,
    _In_z_ const char *rhsText,
    _In_z_ const char *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    return NitsTrue;
}

static NitsResult NITS_CALL CompareA_Checked(
           int lhs,
           int rhs,
    _In_z_ const char *lhsText,
    _In_z_ const char *rhsText,
    _In_z_ const char *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    CheckInjector();
    CompareSwapFunction(NITS.CompareA, CompareA, CompareA_Checked);
    return NITS.CompareA(lhs, rhs, lhsText, rhsText, description, line, mode);
}

static NitsResult NITS_CALL CompareW(
           int lhs,
           int rhs,
    _In_z_ const char *lhsText,
    _In_z_ const char *rhsText,
    _In_z_ const wchar_t *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    return NitsTrue;
}

static NitsResult NITS_CALL CompareW_Checked(
           int lhs,
           int rhs,
    _In_z_ const char *lhsText,
    _In_z_ const char *rhsText,
    _In_z_ const wchar_t *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    CheckInjector();
    CompareSwapFunction(NITS.CompareW, CompareW, CompareW_Checked);
    return NITS.CompareW(lhs, rhs, lhsText, rhsText, description, line, mode);
}

static NitsResult NITS_CALL CompareStringA_(
    _In_z_ const char *lhs,
    _In_z_ const char *rhs,
    _In_z_ const char *lhsText,
    _In_z_ const char *rhsText,
    _In_z_ const char *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    return NitsTrue;
}

static NitsResult NITS_CALL CompareStringA_Checked(
    _In_z_ const char *lhs,
    _In_z_ const char *rhs,
    _In_z_ const char *lhsText,
    _In_z_ const char *rhsText,
    _In_z_ const char *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    CheckInjector();
    CompareSwapFunction(NITS.CompareStringA, CompareStringA_, CompareStringA_Checked);
    return NITS.CompareStringA(lhs, rhs, lhsText, rhsText, description, line, mode);
}

static NitsResult NITS_CALL CompareStringW_(
    _In_z_ const wchar_t *lhs,
    _In_z_ const wchar_t *rhs,
    _In_z_ const char *lhsText,
    _In_z_ const char *rhsText,
    _In_z_ const wchar_t *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    return NitsTrue;
}

static NitsResult NITS_CALL CompareStringW_Checked(
    _In_z_ const wchar_t *lhs,
    _In_z_ const wchar_t *rhs,
    _In_z_ const char *lhsText,
    _In_z_ const char *rhsText,
    _In_z_ const wchar_t *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    CheckInjector();
    CompareSwapFunction(NITS.CompareStringW, CompareStringW_, CompareStringW_Checked);
    return NITS.CompareStringW(lhs, rhs, lhsText, rhsText, description, line, mode);
}

static NitsResult NITS_CALL CompareSubstringA(
    _In_z_ const char *lhs,
    _In_z_ const char *rhs,
    _In_z_ const char *lhsText,
    _In_z_ const char *rhsText,
    _In_z_ const char *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    return NitsTrue;
}

static NitsResult NITS_CALL CompareSubstringA_Checked(
    _In_z_ const char *lhs,
    _In_z_ const char *rhs,
    _In_z_ const char *lhsText,
    _In_z_ const char *rhsText,
    _In_z_ const char *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    CheckInjector();
    CompareSwapFunction(NITS.CompareSubstringA, CompareSubstringA, CompareSubstringA_Checked);
    return NITS.CompareSubstringA(lhs, rhs, lhsText, rhsText, description, line, mode);
}

static NitsResult NITS_CALL CompareSubstringW(
    _In_z_ const wchar_t *lhs,
    _In_z_ const wchar_t *rhs,
    _In_z_ const char *lhsText,
    _In_z_ const char *rhsText,
    _In_z_ const wchar_t *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    return NitsTrue;
}

static NitsResult NITS_CALL CompareSubstringW_Checked(
    _In_z_ const wchar_t *lhs,
    _In_z_ const wchar_t *rhs,
    _In_z_ const char *lhsText,
    _In_z_ const char *rhsText,
    _In_z_ const wchar_t *description,
           NitsCallSite line,
           NitsFaultMode mode)
{
    CheckInjector();
    CompareSwapFunction(NITS.CompareSubstringW, CompareSubstringW, CompareSubstringW_Checked);
    return NITS.CompareSubstringW(lhs, rhs, lhsText, rhsText, description, line, mode);
}

static NitsResult NITS_CALL DidFault()
{
    return NitsFalse;
}

static NitsResult NITS_CALL IsActivated()
{
    return NitsFalse;
}

static NitsResult NITS_CALL IsActivated_Checked()
{
    CheckInjector();
    CompareSwapFunction(NITS.IsActivated, IsActivated, IsActivated_Checked);
    return NITS.IsActivated();
}

static void NITS_CALL StopReportingIgnoredErrors()
{
}

static int *NITS_CALL GetInt(
    _In_z_ const char *name)
{
    return NULL;
}

static int *NITS_CALL GetInt_Checked(
    _In_z_ const char *name)
{
    CheckInjector();
    CompareSwapFunction(NITS.GetInt, GetInt, GetInt_Checked);
    return NITS.GetInt(name);
}

static const char *NITS_CALL GetStringA(
    _In_z_ const char *name)
{
    return NULL;
}

static const char *NITS_CALL GetStringA_Checked(
    _In_z_ const char *name)
{
    CheckInjector();
    CompareSwapFunction(NITS.GetStringA, (void *)GetStringA, (void *)GetStringA_Checked);
    return NITS.GetStringA(name);
}

static const wchar_t *NITS_CALL GetStringW(
    _In_z_ const char *name)
{
    return NULL;
}

static const wchar_t *NITS_CALL GetStringW_Checked(
    _In_z_ const char *name)
{
    CheckInjector();
    CompareSwapFunction(NITS.GetStringW, (void *)GetStringW, (void *)GetStringW_Checked);
    return NITS.GetStringW(name);
}

static void NITS_CALL SetInt(
    _In_z_ const char *name,
           int data)
{
}

static void NITS_CALL SetInt_Checked(
    _In_z_ const char *name,
           int data)
{
    CheckInjector();
    CompareSwapFunction(NITS.SetInt, SetInt, SetInt_Checked);
    NITS.SetInt(name, data);
}

static void NITS_CALL SetStringA(
    _In_z_ const char *name,
    _In_z_ const char *data)
{
}

static void NITS_CALL SetStringA_Checked(
    _In_z_ const char *name,
    _In_z_ const char *data)
{
    CheckInjector();
    CompareSwapFunction(NITS.SetStringA, SetStringA, SetStringA_Checked);
    NITS.SetStringA(name, data);
}

static void NITS_CALL SetStringW(
    _In_z_ const char *name,
    _In_z_ const wchar_t *data)
{
}

static void NITS_CALL SetStringW_Checked(
    _In_z_ const char *name,
    _In_z_ const wchar_t *data)
{
    CheckInjector();
    CompareSwapFunction(NITS.SetStringW, SetStringW, SetStringW_Checked);
    NITS.SetStringW(name, data);
}

static void NITS_CALL SetMode(
    NitsTestMode mode) 
{
}

static void NITS_CALL SetFault(
    _In_opt_z_ const char *function,
    _In_z_     const char *id,
               int attempt)
{
}

static void NITS_CALL SetWait(
    _In_opt_z_ const char *function,
    _In_z_     const char *id,
               int attempt,
    _In_z_     const char *event)
{
}

static void NITS_CALL Signal(
    _In_z_ const char *event)
{
}

static NitsResult NITS_CALL Wait(
    _In_opt_z_ const char *function,
    _In_z_    const char *id,
               int attempt,
               int timeout)
{
    return NitsTrue;
}

static void NITS_CALL ResetFaults()
{
}

static NitsTrapHandle NITS_CALL OpenTrap(
    _In_z_ const char *binary,
    _In_z_ const char *table,
           unsigned version)
{
    return 0;
}

static const void *NITS_CALL GetTrap(
    _In_ NitsTrapHandle handle,
         unsigned offset)
{
    return 0;
}

static const void *NITS_CALL NextTrap(
    _In_z_ const char *table,
           unsigned offset)
{
    return 0;
}

static void *NITS_CALL SetTrap(
    _In_ NitsTrapHandle handle,
    _In_ void *function,
         unsigned offset)
{
    return 0;
}

static void NITS_CALL CopyTrap(
    _In_z_ const char *table,
    _In_z_ const char *from,
    _In_z_ const char *to)
{
}

static void NITS_CALL CloseTrap(
    _In_ NitsTrapHandle handle)
{
}

NITS_DLLEXPORT NitsFT NITS =
{
    ShouldFault_Checked,
    TraceA,
    TraceW,
    AssertA_Checked,
    AssertW_Checked,
    CompareA_Checked,
    CompareW_Checked,
    CompareStringA_Checked,
    CompareStringW_Checked,
    CompareSubstringA_Checked,
    CompareSubstringW_Checked,
    DidFault,
    IsActivated_Checked,
    StopReportingIgnoredErrors,
    GetInt_Checked,
    GetStringA_Checked,
    GetStringW_Checked,
    SetInt_Checked,
    SetStringA_Checked,
    SetStringW_Checked,
    SetMode,
    SetFault,
    SetWait,
    Signal,
    Wait,
    ResetFaults,
    BeginResource_Checked,
    ChangeResource,
    EndResource,
    NULL,
    NULL,
    NULL,
    OpenTrap,
    GetTrap,
    NextTrap,
    SetTrap,
    CopyTrap,
    CloseTrap,
};
