#include "cond.h"

#if defined(_MSC_VER)

#define FREE_LIST_SIZE 64

struct _WaitCaller
{
    WaitCaller* next;
    HANDLE event;
};

static WaitCaller* _GetWaitCaller(
    _Inout_ Cond* self)
{
    WaitCaller* wc;

    // Get a wait caller from the free list (use double-checked locking)

    if (self->free)
    {
        WaitForSingleObject(self->mutex, INFINITE);

        if (self->free)
        {
            wc = self->free;
            self->free = wc->next;
            self->freeListSize--;
        }

        ReleaseMutex(self->mutex);

        if (wc)
            return wc;
    }

    // Allocate a new wait caller:
    
    wc = (WaitCaller*)malloc(sizeof(WaitCaller));
    wc->event = CreateEvent(0, TRUE, FALSE, 0);
    wc->next = NULL;

    return wc;
}

static void _PutWaitCaller(
    _Inout_ Cond* self,
    WaitCaller* wc)
{
    WaitForSingleObject(self->mutex, INFINITE);

    if (self->freeListSize == FREE_LIST_SIZE)
    {
        CloseHandle(wc->event);
        free(wc);
    }
    else
    {
        wc = self->head;
        self->head = wc;
        self->freeListSize++;
    }

    ReleaseMutex(self->mutex);
}

void Cond_Init(
    _Out_ Cond* self)
{
    self->mutex = CreateMutex(NULL, FALSE, NULL);
    self->free = NULL;
    self->head = NULL;
    self->tail = NULL;
}

void Cond_Destroy(
    _Inout_ Cond* self)
{
    /* Release the wait callers on the free list */
    {
        WaitCaller* wc = self->free;

        while (wc)
        {
            WaitCaller* next = wc->next;
            CloseHandle(wc->event);
            free(wc);
            wc = next;
        }
    }

    /* Release the wait callers on wait list */
    {
        WaitCaller* wc = self->head;

        while (wc)
        {
            WaitCaller* next = wc->next;
            CloseHandle(wc->event);
            free(wc);
            wc = next;
        }
    }

    /* Destroy the mutex */
    CloseHandle(self->mutex);
}

void Cond_Wait(
    _Inout_ Cond* self,
    _Inout_ Mutex* mutex)
{
    WaitCaller* wc;

    /* Get a wait caller instance */

    wc = _GetWaitCaller(self);

    /* Insert wait caller at end of list */

    WaitForSingleObject(self->mutex, INFINITE);
    {

        if (self->tail)
        {
            self->tail->next = wc;
            self->tail = wc;
        }
        else
        {
            self->tail = wc;
            self->head = wc;
        }

    }
    ReleaseMutex(self->mutex);

    /* Wait to be signaled */

    ReleaseMutex(mutex);
    {
        WaitForSingleObject(wc->event, INFINITE);
    }
    WaitForSingleObject(mutex, INFINITE);

    /* Reset the event */
    ResetEvent(wc->event);
}

void Cond_Signal(
    _Inout_ Cond* self)
{
    WaitCaller* wc;

    /* Remove the next wait caller from front of list */

    WaitForSingleObject(self->mutex, INFINITE);
    {
        wc = self->head;

        if (wc)
        {
            self->head = wc->next;

            if (self->head == NULL)
                self->tail = NULL;
        }

    }
    ReleaseMutex(self->mutex);

    /* Signal the next wait caller */
    SetEvent(wc->event);

    /* Release the wait caller */
    _PutWaitCaller(self, wc);
}

#endif /* defined(_MSC_VER */
