#include "shmem.h"
#include <stdio.h>

#if defined(PAL_HAVE_POSIX)
# include <unistd.h>
# include <sys/mman.h>
# include <sys/shm.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <string.h>
#endif

int Shmem_Open(
    _Out_ Shmem* self,
    _In_ const PAL_Char* name,
    _In_ ShmemAccess access,
    _In_ size_t size)
{
#if defined(_MSC_VER)

    DWORD protect = 0;
    DWORD hi;
    DWORD lo;

    if (access == SHMEM_ACCESS_READWRITE)
        protect = PAGE_READWRITE;
    else if (access == SHMEM_ACCESS_READONLY)
        protect = PAGE_READONLY;
    else
        return -1;

    hi = (DWORD)((0xFFFFFFFF00000000 & (PAL_Uint64)size) >> 32);
    lo = (DWORD)(0x00000000FFFFFFFF & (PAL_Uint64)size);

    self->handle = CreateFileMapping( 
        INVALID_HANDLE_VALUE,   /* use paging file */
        NULL,                   /* default security attributes */
        protect,                /* page access rights */
        hi,                     /* size: high 32-bits */
        lo,                     /* size: low 32-bits */
        name);                  /* name of map object */

    return self->handle ? 0 : -1;

#else

    int oflag;

    /* Build the oflag */

    if (access == SHMEM_ACCESS_READWRITE)
        oflag = O_CREAT | O_RDWR;
    else if (access == SHMEM_ACCESS_READONLY)
        oflag = O_CREAT | O_RDONLY;
    else
        return -1;

    /* Open the shared memory segment */

    if ((self->shmid = shm_open(name, oflag, S_IRUSR | S_IWUSR)) == -1)
    {
        return -1;
    }

    if (ftruncate(self->shmid, size) != 0)
    {
        close(self->shmid);
        return -1;
    }

    return 0;

#endif
}

int Shmem_Close(
    _Inout_ Shmem* self)
{
#if defined(_MSC_VER)
    return CloseHandle(self->handle) ? 0 : -1;
#else
    return close(self->shmid) == 0 ? 0 : -1;
#endif
}

void* Shmem_Map(
    _Inout_ Shmem* self,
    _In_ ShmemAccess access,
    _In_ ptrdiff_t offset,
    _In_ size_t size)
{
#if defined(_MSC_VER)

    DWORD access = 0;
    DWORD lo;
    DWORD hi;
    void* ptr;

    if (access == SHMEM_ACCESS_READWRITE)
        access = FILE_MAP_ALL_ACCESS;
    else if (access == SHMEM_ACCESS_READONLY)
        access = FILE_MAP_READ;
    else
        return NULL;

    lo = (DWORD)(0x00000000FFFFFFFF & (PAL_Uint64)offset);
    hi = (DWORD)((0xFFFFFFFF00000000 & (PAL_Uint64)offset) >> 32);

    ptr = MapViewOfFile(
        self->handle,
        access,
        hi, /* offset: high 32-bits */
        lo, /* offset: low 32-bits */
        size);

    return ptr;

#else

    int mprot;
    int mflags;
    void* ptr;

    if (access == SHMEM_ACCESS_READWRITE)
        mprot = PROT_READ | PROT_WRITE;
    else if (access == SHMEM_ACCESS_READONLY)
        mprot = PROT_READ;
    else
        return NULL;

    mflags = MAP_FILE | MAP_SHARED;

    ptr = mmap(NULL, size, mprot, mflags, self->shmid, offset);

    if (ptr == MAP_FAILED)
        return NULL;

    return ptr;

#endif
}

int Shmem_Unmap(
    _Inout_ Shmem* self,
    _In_ void* ptr,
    _In_ size_t size)
{
#if defined(_MSC_VER)
    return UnmapViewOfFile(ptr) ? 0 : -1;
#else
    int r;
    
    if ((r = munmap(ptr, size)) == -1)
        return -1;

    return 0;
#endif
}

size_t Shmem_Granularity()
{
#if defined(_MSC_VER)
    SYSTEM_INFO info;
    GetSystemInfo(&info);
    return info.dwAllocationGranularity;
#else
    return sysconf(_SC_PAGESIZE);
#endif
}
