/*
**==============================================================================
**
** strings.h:
**
**     This file defines string-manipluation functions. For each string 
**     operation, there are three implementations:
**
**         - Single-character - char
**         - Wide-character - wchar_t
**         - T-character - TChar (char or wchar_t depending on configuration)
**
**     Each string function has one of the following prefixes:
**
**         - Str - single character string
**         - Wcs - wide character string
**         - Tcs - T-character string
**
**     For example, the are three variants for simple copying of strings:
**
**         - Strcpy() - same as strcpy()
**         - Wcscpy() - same as wcscpy()
**         - Tcscpy() - same as strcpy() or wcscpy() depending on configuration.
**
**     When passing string literals to the Tcs* variant functions, use the
**     PAL_T() macro. For example:
**
**         TChar buf[32];
**         Tcscpy(buf, PAL_T("Hello"));
**
**     For the typedef TChar, one of the following is true:
**
**         sizeof(TChar) == sizeof(char)
**         sizeof(TChar) == sizeof(wchar_t)
**
**     If the latter is true, then PAL_USE_WCHAR is defined.
**
**==============================================================================
*/

#ifndef _pal_strings_h
#define _pal_strings_h

#include <wchar.h>
#include <string.h>
#include "palcommon.h"

#undef PAL_USE_WCHAR

PAL_BEGIN_EXTERNC

typedef PAL_Char TChar;

/*
**==============================================================================
**
** Strcmp()
** Wcscmp()
** Tcscmp()
**
**==============================================================================
*/

PAL_INLINE int Wcscmp(const wchar_t* s1, const wchar_t* s2)
{
    return wcscmp(s1, s2);
}

PAL_INLINE int Strcmp(const char* s1, const char* s2)
{
    return strcmp(s1, s2);
}

PAL_INLINE int Tcscmp(const TChar* s1, const TChar* s2)
{
#if defined(PAL_USE_WCHAR)
    return Wcscmp(s1, s2);
#else
    return Strcmp(s1, s2);
#endif
}

/*
**==============================================================================
**
** Strncmp()
** Wcsncmp()
** Tcsncmp()
**
**==============================================================================
*/

PAL_INLINE int Strncmp(const char* s1, const char* s2, size_t n)
{
    return strncmp(s1, s2, n);
}

PAL_INLINE int Wcsncmp(const wchar_t* s1, const wchar_t* s2, size_t n)
{
    return wcsncmp(s1, s2, n);
}

PAL_INLINE int Tcsncmp(const TChar* s1, const TChar* s2, size_t n)
{
#if defined(PAL_USE_WCHAR)
    return Wcsncmp(s1, s2, n);
#else
    return Strncmp(s1, s2, n);
#endif
}

/*
**==============================================================================
**
** Strlen()
** Wcslen()
** Tcslen()
**
**==============================================================================
*/

PAL_INLINE size_t Strlen(const char* str)
{
    return strlen(str);
}

PAL_INLINE size_t Wcslen(const wchar_t* str)
{
    return wcslen(str);
}

PAL_INLINE size_t Tcslen(const TChar* str)
{
#if defined(PAL_USE_WCHAR)
    return Wcslen(str);
#else
    return Strlen(str);
#endif
}

/*
**==============================================================================
**
** Strdup()
** Wcsdup()
** Tcsdup()
**
**==============================================================================
*/

PAL_INLINE char* Strdup(const char* s)
{
#if defined(_MSC_VER)
    return _strdup(s);
#else
    return strdup(s);
#endif
}

PAL_INLINE wchar_t* Wcsdup(const wchar_t* s)
{
#if defined(_MSC_VER)
    return _wcsdup(s);
#else
    extern wchar_t* wcsdup(const wchar_t*);
    return wcsdup(s);
#endif
}

PAL_INLINE TChar* Tcsdup(const TChar* s)
{
#if defined(PAL_USE_WCHAR)
    return Wcsdup(s);
#else
    return Strdup(s);
#endif
}

/*
**==============================================================================
**
** Strcasecmp()
** Wcscasecmp()
** Tcscasecmp()
**
**==============================================================================
*/

PAL_INLINE int Strcasecmp(const char* s1, const char* s2)
{
#if defined(_MSC_VER)
    return _stricmp(s1, s2);
#else
    return strcasecmp(s1, s2);
#endif
}

PAL_INLINE int Wcscasecmp(const wchar_t* s1, const wchar_t* s2)
{
#if defined(_MSC_VER)
    return _wcsicmp(s1, s2);
#else
    extern int wcscasecmp(const wchar_t*, const wchar_t*);
    return wcscasecmp(s1, s2);
#endif
}

PAL_INLINE int Tcscasecmp(const TChar* s1, const TChar* s2)
{
#if defined(PAL_USE_WCHAR)
    return Wcscasecmp(s1, s2);
#else
    return Strcasecmp(s1, s2);
#endif
}

/*
**==============================================================================
**
** Strncasecmp()
** Wcsncasecmp()
** Tcsncasecmp()
**
**==============================================================================
*/

PAL_INLINE int Strncasecmp(const char* s1, const char* s2, size_t n)
{
#if defined(_MSC_VER)
    return _strnicmp(s1, s2, n);
#else
    return strncasecmp(s1, s2, n);
#endif
}

PAL_INLINE int Wcsncasecmp(const wchar_t* s1, const wchar_t* s2, size_t n)
{
#if defined(_MSC_VER)
    return _wcsnicmp(s1, s2, n);
#else
    extern int wcsncasecmp(const wchar_t*, const wchar_t*, size_t n);
    return wcsncasecmp(s1, s2, n);
#endif
}

PAL_INLINE int Tcsncasecmp(const TChar* s1, const TChar* s2, size_t n)
{
#if defined(PAL_USE_WCHAR)
    return Wcsncasecmp(s1, s2, n);
#else
    return Strncasecmp(s1, s2, n);
#endif
}

/*
**==============================================================================
**
** Strtoul()
** Wcstoul()
** Tcstoul()
**
**==============================================================================
*/

PAL_INLINE unsigned long Strtoul(const char* str, char** end, int base)
{
    return strtoul(str, end, base);
}

PAL_INLINE unsigned long Wcstoul(const wchar_t* str, wchar_t** end, int base)
{
    return wcstoul(str, end, base);
}

PAL_INLINE unsigned long Tcstoul(const TChar* str, TChar** end, int base)
{
#if defined(PAL_USE_WCHAR)
    return Wcstoul(str, end, base);
#else
    return Strtoul(str, end, base);
#endif
}

/*
**==============================================================================
**
** Strtol()
** Wcstol()
** Tcstol()
**
**==============================================================================
*/

PAL_INLINE long Strtol(const char* str, char** end, int base)
{
    return strtol(str, end, base);
}

PAL_INLINE long Wcstol(const wchar_t* str, wchar_t** end, int base)
{
    return wcstol(str, end, base);
}

PAL_INLINE long Tcstol(const TChar* str, TChar** end, int base)
{
#if defined(PAL_USE_WCHAR)
    return Wcstol(str, end, base);
#else
    return Strtol(str, end, base);
#endif
}

/*
**==============================================================================
**
** Strtoull()
** Wcstoull()
** Tcstoull()
**
**==============================================================================
*/

PAL_INLINE PAL_Uint64 Strtoull(const char* str, char** end, int base)
{
#if defined(_MSC_VER)
    return _strtoui64(str, end, base);
#else
    return strtoull(str, end, base);
#endif
}

#if defined(PAL_USE_WCHAR)
PAL_INLINE PAL_Uint64 Wcstoull(const wchar_t* str, wchar_t** end, int base)
{
#if defined(_MSC_VER)
    return _wcstoui64(str, end, base);
#else
    extern unsigned long long wcstoull(const wchar_t* s, wchar_t** e, int b);
    return wcstoull(str, end, base);
#endif
}
#endif

PAL_INLINE PAL_Uint64 Tcstoull(const TChar* str, TChar** end, int base)
{
#if defined(PAL_USE_WCHAR)
    return Wcstoull(str, end, base);
#else
    return Strtoull(str, end, base);
#endif
}

/*
**==============================================================================
**
** Strtoll()
** Wcstoll()
** Tcstoll()
**
**==============================================================================
*/

PAL_INLINE PAL_Sint64 Strtoll(const char* str, char** end, int base)
{
#if defined(_MSC_VER)
    return _strtoi64(str, end, base);
#else
    return strtoll(str, end, base);
#endif
}

PAL_INLINE PAL_Sint64 Wcstoll(const wchar_t* str, wchar_t** end, int base)
{
#if defined(_MSC_VER)
    return _wcstoi64(str, end, base);
#else
    extern long long wcstoll(const wchar_t* s, wchar_t** e, int b);
    return wcstoll(str, end, base);
#endif
}

PAL_INLINE PAL_Sint64 Tcstoll(const TChar* str, TChar** end, int base)
{
#if defined(PAL_USE_WCHAR)
    return Wcstoll(str, end, base);
#else
    return Strtoll(str, end, base);
#endif
}

/*
**==============================================================================
**
** Strtod()
** Wcstod()
** Tcstod()
**
**==============================================================================
*/

PAL_INLINE double Strtod(const char* str, char** end)
{
    return strtod(str, end);
}

PAL_INLINE double Wcstod(const wchar_t* str, wchar_t** end)
{
    return wcstod(str, end);
}

PAL_INLINE double Tcstod(const TChar* str, TChar** end)
{
#if defined(PAL_USE_WCHAR)
    return Wcstod(str, end);
#else
    return Strtod(str, end);
#endif
}

/*
**==============================================================================
**
** Strchr()
** Wcschr()
** Tcschr()
**
**==============================================================================
*/

PAL_INLINE char* Strchr(const char* str, char ch)
{
    return (char*)strchr(str, ch);
}

PAL_INLINE wchar_t* Wcschr(const wchar_t* str, wchar_t ch)
{
    return (wchar_t*)wcschr(str, ch);
}

PAL_INLINE TChar* Tcschr(const TChar* str, TChar ch)
{
#if defined(PAL_USE_WCHAR)
    return (wchar_t*)Wcschr(str, ch);
#else
    return (char*)Strchr(str, ch);
#endif
}

/*
**==============================================================================
**
** Strrchr()
** Wcsrchr()
** Tcsrchr()
**
**==============================================================================
*/

PAL_INLINE char* Strrchr(const char* str, char ch)
{
    return (char*)strrchr(str, ch);
}

PAL_INLINE wchar_t* Wcsrchr(const wchar_t* str, wchar_t ch)
{
    return (wchar_t*)wcsrchr(str, ch);
}

PAL_INLINE TChar* Tcsrchr(const TChar* str, TChar ch)
{
#if defined(PAL_USE_WCHAR)
    return Wcsrchr(str, ch);
#else
    return Strrchr(str, ch);
#endif
}

/*
**==============================================================================
**
** Strcat()
** Wcscat()
** Tcscat()
**
**==============================================================================
*/

PAL_INLINE char* Strcat(char* dest, size_t count, const char* src)
{
#if defined(_MSC_VER)
    strcat_s(dest, count, src);
    return dest;
#else
    return strcat(dest, src);
#endif
}

PAL_INLINE wchar_t* Wcscat(wchar_t* dest, size_t count, const wchar_t* src)
{
#if defined(_MSC_VER)
    wcscat_s(dest, count, src);
    return dest;
#else
    return wcscat(dest, src);
#endif
}

PAL_INLINE TChar* Tcscat(TChar* dest, size_t count, const TChar* src)
{
#if defined(PAL_USE_WCHAR)
    return Wcscat(dest, count, src);
#else
    return Strcat(dest, count, src);
#endif
}

/*
**==============================================================================
**
** Strtok()
** Wcstok()
** Tcstok()
**
**==============================================================================
*/

PAL_INLINE char* Strtok(char* str, const char* delim, char** ctx)
{
#if defined(_MSC_VER)
    return strtok_s(str, delim, ctx);
#else
    return strtok_r(str, delim, ctx);
#endif
}

PAL_INLINE wchar_t* Wcstok(wchar_t* str, const wchar_t* delim, wchar_t** ctx)
{
#if defined(_MSC_VER)
    return wcstok_s(str, delim, ctx);
#else
    return wcstok(str, delim, ctx);
#endif
}

PAL_INLINE TChar* Tcstok(TChar* str, const TChar* delim, TChar** ctx)
{
#if defined(PAL_USE_WCHAR)
    return Wcstok(str, delim, ctx);
#else
    return Strtok(str, delim, ctx);
#endif
}

/*
**==============================================================================
**
** Strlcat()
** Wcslcat()
** Tcslcat()
**
**==============================================================================
*/

size_t Strlcat(char* dest, const char* src, size_t size);

size_t Wcslcat(wchar_t* dest, const wchar_t* src, size_t size);

PAL_INLINE size_t Tcslcat(TChar* dest, const TChar* src, size_t size)
{
#if defined(PAL_USE_WCHAR)
    return Wcslcat(dest, src, size);
#else
    return Strlcat(dest, src, size);
#endif
}

/*
**==============================================================================
**
** Strlcpy()
** Wcslcpy()
** Tcslcpy()
**
**==============================================================================
*/

size_t Strlcpy(char* dest, const char* src, size_t size);

size_t Wcslcpy(wchar_t* dest, const wchar_t* src, size_t size);

PAL_INLINE size_t Tcslcpy(TChar* dest, const TChar* src, size_t size)
{
#if defined(PAL_USE_WCHAR)
    return Wcslcpy(dest, src, size);
#else
    return Strlcpy(dest, src, size);
#endif
}

PAL_END_EXTERNC

#endif /* _pal_strings_h */
