#include <vector>
#include <string>
#include <algorithm>
#include <cassert>
#include <cstdio>
#include <pal/atomic.h>
#include <pal/thread.h>
#include <pal/tls.h>
#include <pal/sleep.h>
#include <pal/once.h>
#include <pal/mutex.h>
#include <pal/cond.h>
#include <pal/cpu.h>
#include <pal/shlib.h>
#include <pal/slist.h>
#include <pal/sem.h>
#include <pal/shmem.h>
#include <pal/strings.h>
#include <pal/dir.h>

#define TEST_ASSERT assert

using namespace std;

//==============================================================================
//
// TestAtomic()
//
//==============================================================================

static void TestAtomic()
{
    // Atomic_Inc()
    {
        ptrdiff_t x = 0;
        ptrdiff_t r = Atomic_Inc(&x);
        TEST_ASSERT(r == 1);
        TEST_ASSERT(x == 1);
        r = Atomic_Inc(&x);
        TEST_ASSERT(r == 2);
        TEST_ASSERT(x == 2);
    }

    // Atomic_Dec()
    {
        ptrdiff_t x = 2;
        ptrdiff_t r = Atomic_Dec(&x);
        TEST_ASSERT(r == 1);
        TEST_ASSERT(x == 1);
        r = Atomic_Dec(&x);
        TEST_ASSERT(r == 0);
        TEST_ASSERT(x == 0);
    }

    // Atomic_CompareAndSwap()
    {
        ptrdiff_t x = 2;
        ptrdiff_t r = Atomic_CompareAndSwap(&x, 2, 10); 
        TEST_ASSERT(r == 2);
        TEST_ASSERT(x == 10);
    }

    // Atomic_CompareAndSwap()
    {
        ptrdiff_t x = 2;
        ptrdiff_t r = Atomic_CompareAndSwap(&x, 500, 10); 
        TEST_ASSERT(r == 2);
        TEST_ASSERT(x == 2);
    }

    // Atomic_Swap()
    {
        ptrdiff_t x = 100;
        ptrdiff_t r = Atomic_Swap(&x, 200);
        TEST_ASSERT(r == 100);
        TEST_ASSERT(x == 200);
    }

    // Atomic_Lock() / Atomic_Unlock()
    {
        ptrdiff_t lock = 0;

        Atomic_Lock(&lock);
        Atomic_Unlock(&lock);
    }
}

//==============================================================================
//
// TestThread()
//
//==============================================================================

static PAL_Uint32 THREAD_API TestThread1(void* arg)
{
    const char* name = (const char*)arg;
    TEST_ASSERT(strcmp(name, "TestThread1") == 0);

    return 1001;
}

static PAL_Uint32 THREAD_API TestThread2(void* arg)
{
    const char* name = (const char*)arg;
    TEST_ASSERT(strcmp(name, "TestThread2") == 0);

    return 1002;
}

static void TestThread()
{
    Thread t1;

    if (Thread_CreateJoinable(&t1, TestThread1, (char*)"TestThread1") != 0)
        TEST_ASSERT(false);

    Thread t2;

    if (Thread_CreateJoinable(&t2, TestThread2, (char*)"TestThread2") != 0)
        TEST_ASSERT(false);

    PAL_Uint32 arg1;
    Thread_Join(&t1, &arg1);
    Thread_Destroy(&t1);
    TEST_ASSERT(arg1 == 1001);

    PAL_Uint32 arg2;
    Thread_Join(&t2, &arg2);
    Thread_Destroy(&t2);
    TEST_ASSERT(arg2 == 1002);
}

//==============================================================================
//
// TestTLS()
//
//==============================================================================

static PAL_Uint32 THREAD_API TestTLS1(void* arg)
{
    TLS* tls = (TLS*)arg;

    for (ptrdiff_t i = 0; i < 100; i++)
    {
        TLS_Set(tls, i);
        Sleep_Milliseconds(5);
        ptrdiff_t x = TLS_Get(tls);
        TEST_ASSERT(x == i);
    }

    return NULL;
}

static PAL_Uint32 THREAD_API TestTLS2(void* arg)
{
    TLS* tls = (TLS*)arg;

    for (ptrdiff_t i = 100; i < 200; i++)
    {
        TLS_Set(tls, i);
        Sleep_Milliseconds(5);
        ptrdiff_t x = TLS_Get(tls);
        TEST_ASSERT(x == i);
    }

    return NULL;
}

static void TestTLS()
{
    Thread t1;
    TLS tls;

    int r = TLS_Init(&tls);
    TEST_ASSERT(r == 0);

    if (Thread_CreateJoinable(&t1, TestTLS1, &tls) != 0)
        TEST_ASSERT(false);

    Thread t2;

    if (Thread_CreateJoinable(&t2, TestTLS2, &tls) != 0)
        TEST_ASSERT(false);

    Sleep_Milliseconds(10);

    PAL_Uint32 arg1;
    Thread_Join(&t1, &arg1);
    Thread_Destroy(&t1);

    PAL_Uint32 arg2;
    Thread_Join(&t2, &arg2);
    Thread_Destroy(&t2);

    TLS_Destroy(&tls);
}

//==============================================================================
//
// TestOnce()
//
//==============================================================================

static Once s_once = ONCE_INITIALIZER;

void IncrementCount(void* arg)
{
    size_t* countPtr = (size_t*)arg;
    (*countPtr)++;
}

static PAL_Uint32 THREAD_API TestOnce1(void* arg)
{
    size_t* countPtr = (size_t*)arg;

    for (ptrdiff_t i = 0; i < 100; i++)
    {
        Sleep_Milliseconds(5);
        Once_Invoke(&s_once, IncrementCount, countPtr);
    }

    return 0;
}

static PAL_Uint32 THREAD_API TestOnce2(void* arg)
{
    size_t* countPtr = (size_t*)arg;

    for (ptrdiff_t i = 0; i < 100; i++)
    {
        Sleep_Milliseconds(5);
        Once_Invoke(&s_once, IncrementCount, countPtr);
    }

    return 0;
}

static void TestOnce()
{
    Thread t1;
    size_t count = 0;

    if (Thread_CreateJoinable(&t1, TestOnce1, &count) != 0)
        TEST_ASSERT(false);

    Thread t2;

    if (Thread_CreateJoinable(&t2, TestOnce2, &count) != 0)
        TEST_ASSERT(false);

    Sleep_Milliseconds(10);

    PAL_Uint32 arg1;
    Thread_Join(&t1, &arg1);
    Thread_Destroy(&t1);

    PAL_Uint32 arg2;
    Thread_Join(&t2, &arg2);
    Thread_Destroy(&t2);

    TEST_ASSERT(count == 1);
}

//==============================================================================
//
// TestMutex()
//
//==============================================================================

static Mutex s_mutex = MUTEX_INITIALIZER;

static PAL_Uint32 THREAD_API TestMutex1(void* arg)
{
    size_t* countPtr = (size_t*)arg;

    for (ptrdiff_t i = 0; i < 100; i++)
    {
        Mutex_Acquire(&s_mutex);
        size_t count = *countPtr;
        Sleep_Milliseconds(1);
        count++;
        Sleep_Milliseconds(1);
        *countPtr = count;
        Mutex_Release(&s_mutex);
    }

    return NULL;
}

static PAL_Uint32 THREAD_API TestMutex2(void* arg)
{
    size_t* countPtr = (size_t*)arg;

    for (ptrdiff_t i = 0; i < 100; i++)
    {
        Mutex_Acquire(&s_mutex);
        size_t count = *countPtr;
        Sleep_Milliseconds(1);
        count++;
        Sleep_Milliseconds(1);
        *countPtr = count;
        Mutex_Release(&s_mutex);
    }

    return NULL;
}

static void TestMutex()
{
    Thread t1;
    size_t count = 0;

    if (Thread_CreateJoinable(&t1, TestMutex1, &count) != 0)
        TEST_ASSERT(false);

    Thread t2;

    if (Thread_CreateJoinable(&t2, TestMutex2, &count) != 0)
        TEST_ASSERT(false);

    Sleep_Milliseconds(10);

    PAL_Uint32 arg1;
    Thread_Join(&t1, &arg1);
    Thread_Destroy(&t1);

    PAL_Uint32 arg2;
    Thread_Join(&t2, &arg2);
    Thread_Destroy(&t2);

    TEST_ASSERT(count == 200);
}

//==============================================================================
//
// TestCond()
//
//==============================================================================

#if 0

class Queue
{
public:

    Queue()
    {
        Mutex_Init(&m_mutex);
        Cond_Init(&m_notEmpty);
        Cond_Init(&m_notFull);
    }

    ~Queue()
    {
        Mutex_Destroy(&m_mutex);
        Cond_Destroy(&m_notEmpty);
        Cond_Destroy(&m_notFull);
    }

    void Put(unsigned int x)
    {
        Mutex_Acquire(&m_mutex);
        {
            while (m_data.size() == SIZE)
                Cond_Wait(&m_notFull, &m_mutex);

            m_data.push_back(x);
            Cond_Signal(&m_notEmpty);
        }
        Mutex_Release(&m_mutex);
    }

    void Get(unsigned int& x)
    {
        Mutex_Acquire(&m_mutex);
        {
            while (m_data.size() == 0)
                Cond_Wait(&m_notEmpty, &m_mutex);

            x = m_data.front();
            m_data.pop_front();
            Cond_Signal(&m_notFull);
        }
        Mutex_Release(&m_mutex);
    }

private:

    enum { SIZE = 100 };
    Mutex m_mutex;
    Cond m_notEmpty;
    Cond m_notFull;
    list<unsigned int> m_data;
};

static const size_t NREADERS = 20;
static const size_t NPUTS = 10000;

static ptrdiff_t s_sum = 0;
static ptrdiff_t s_sumLock = 0;

static void* TestCondReader(void* arg)
{
    Queue* queue = (Queue*)arg;

    for (;;)
    {
        unsigned int x;

        queue->Get(x);

        if (x == 0)
            break;

        Atomic_Lock(&s_sumLock);
        s_sum += x;
        Atomic_Unlock(&s_sumLock);
    }

    return NULL;
}

static void* TestCondWriter(void* arg)
{
    Queue* queue = (Queue*)arg;

    for (unsigned int i = 1; i <= NPUTS; i++)
    {
        queue->Put(i);
    }

    // Stop all the readers:

    for (unsigned int i = 0; i < NREADERS; i++)
    {
        queue->Put(0);
    }

    return NULL;
}

static void TestCond()
{
    // Create the queue:
    
    Queue queue;
    
    // Create writer thread:
    
    Thread t1;

    if (Thread_CreateJoinable(&t1, TestCondWriter, &queue) != 0)
        TEST_ASSERT(false);

    // Create reader threads:

    vector<Thread> readers;

    for (size_t i = 0; i < NREADERS; i++)
    {
        Thread t;

        if (Thread_CreateJoinable(&t, TestCondReader, &queue) != 0)
            TEST_ASSERT(false);

        readers.push_back(t);
    }

    // Join with writer:

    void* arg1;
    Thread_Join(&t1, &arg1);
    Thread_Destroy(&t1);

    // Join with readers:

    for (size_t i = 0; i < NREADERS; i++)
    {
        void* arg2;
        Thread_Join(&readers[i], &arg2);
        Thread_Destroy(&readers[i]);
    }

    // Verify that the sum is right:
    
    ptrdiff_t sum = 0;

    for (unsigned int i = 1; i <= NPUTS; i++)
        sum += i;

    TEST_ASSERT(sum == s_sum);
}

#endif

//==============================================================================
//
// TestCPU()
//
//==============================================================================

static void TestCPU()
{
    int count = CPU_GetCount();
    TEST_ASSERT(count >= 1);

    int current = CPU_GetCurrent();
    TEST_ASSERT(current >= 0);

    if (count == 1)
        TEST_ASSERT(current == 0);
}

//==============================================================================
//
// TestShlib()
//
//==============================================================================

static void TestShlib()
{
// ATTN:
#if 0
    Shlib* shlib = Shlib_Open(CONFIG_LIBDIR "/libtestshlib.so");
    TEST_ASSERT(shlib != NULL);

    typedef const char* (*Function)();
    Function func = (Function)Shlib_Sym(shlib, "TestShlib");
    TEST_ASSERT(func != NULL);

    const char* str = (*func)();
    TEST_ASSERT(strcmp(str, "TestShlib") == 0);
#endif
}

//==============================================================================
//
// TestSList()
//
//==============================================================================

static const int NINTEGERS = 10000;

struct IntegerEntry
{
    SListEntry entry;
    int x;
};

static PAL_Uint32 THREAD_API TestListPush(void* arg)
{
    SListHead* head = (SListHead*)arg;

    for (int i = 0; i < NINTEGERS; i++)
    {
        IntegerEntry* p = new IntegerEntry;
        p->x = i;
        SList_PushAtomic(head, &p->entry);
    }

    return 0;
}

static PAL_Uint32 THREAD_API TestListPop(void* arg)
{
    SListHead* head = (SListHead*)arg;

    for (int i = 0; i < NINTEGERS; i++)
    {
        IntegerEntry* p = (IntegerEntry*)SList_PopAtomic(head);
        TEST_ASSERT(p != NULL);
    }

    return 0;
}

static void TestSList()
{
    /* Test push */
    {
        Thread t1;
        SListHead head;

        SList_Init(&head);

        if (Thread_CreateJoinable(&t1, TestListPush, &head) != 0)
            TEST_ASSERT(false);

        Thread t2;

        if (Thread_CreateJoinable(&t2, TestListPush, &head) != 0)
            TEST_ASSERT(false);

        PAL_Uint32 arg1;
        Thread_Join(&t1, &arg1);
        Thread_Destroy(&t1);

        PAL_Uint32 arg2;
        Thread_Join(&t2, &arg2);
        Thread_Destroy(&t2);

        // Check the list:
        
        int expect = 0;

        for (int i = 0; i < NINTEGERS; i++)
        {
            expect += i;
            expect += i;
        }

        SListEntry* first = SList_FlushAtomic(&head);

        int total = 0;
        size_t size = 0;

        for (SListEntry* p = first; p; )
        {
            SListEntry* next = SList_Next(p);
            IntegerEntry* entry = (IntegerEntry*)p;
            total += entry->x;
            delete entry;
            p = next;
            size++;
        }

        TEST_ASSERT(size == 2 * NINTEGERS);
        TEST_ASSERT(expect == total);
    }

    /* Test pop */
    {
        SListHead head;
        SList_Init(&head);

        // Push phase:
        {
            Thread t1;

            if (Thread_CreateJoinable(&t1, TestListPush, &head) != 0)
                TEST_ASSERT(false);

            Thread t2;

            if (Thread_CreateJoinable(&t2, TestListPush, &head) != 0)
                TEST_ASSERT(false);

            PAL_Uint32 arg1;
            Thread_Join(&t1, &arg1);
            Thread_Destroy(&t1);

            PAL_Uint32 arg2;
            Thread_Join(&t2, &arg2);
            Thread_Destroy(&t2);
        }

        // Pop phase:
        {
            Thread t1;

            if (Thread_CreateJoinable(&t1, TestListPop, &head) != 0)
                TEST_ASSERT(false);

            Thread t2;

            if (Thread_CreateJoinable(&t2, TestListPop, &head) != 0)
                TEST_ASSERT(false);

            PAL_Uint32 arg1;
            Thread_Join(&t1, &arg1);
            Thread_Destroy(&t1);

            PAL_Uint32 arg2;
            Thread_Join(&t2, &arg2);
            Thread_Destroy(&t2);
        }

        // Check that the list is empty:
        SListEntry* first = SList_FlushAtomic(&head);
        TEST_ASSERT(first == NULL);
    }
}

//==============================================================================
//
// TestSem()
//
//==============================================================================

struct TestSemData
{
    Sem sem;
    unsigned int sum;
    unsigned int count;
};

static PAL_Uint32 THREAD_API TestSemThread(void* arg)
{
    TestSemData* data = (TestSemData*)arg;

    for (unsigned int i = 0; i < data->count; i++)
    {
        Sem_Wait(&data->sem);
        size_t tmp = data->sum;
        Sleep_Milliseconds(1);
        tmp++;
        Sleep_Milliseconds(1);
        data->sum = tmp;
        Sleep_Milliseconds(1);
        Sem_Post(&data->sem, 1);
    }

    return 0;
}

static void TestSem()
{
    Thread t1;
    TestSemData data;

    data.sum = 0;
    data.count = 0;

    int r = Sem_Init(&data.sem, 1);
    TEST_ASSERT(r == 0);

    if (Thread_CreateJoinable(&t1, TestSemThread, &data) != 0)
        TEST_ASSERT(false);

    Thread t2;

    if (Thread_CreateJoinable(&t2, TestSemThread, &data) != 0)
        TEST_ASSERT(false);

    PAL_Uint32 arg1;
    Thread_Join(&t1, &arg1);
    Thread_Destroy(&t1);

    PAL_Uint32 arg2;
    Thread_Join(&t2, &arg2);
    Thread_Destroy(&t2);

    Sem_Destroy(&data.sem);


    // Check the result:
    
    TEST_ASSERT(data.sum == 2 * data.count);
}

//==============================================================================
//
// TestNamedSem()
//
//==============================================================================

#if defined(_MSC_VER)
# define SEMAPHORE_NAME L"/PALTestSem"
#else
# define SEMAPHORE_NAME "/PALTestSem"
#endif

struct TestNamedSemData
{
    unsigned int sum;
    unsigned int count;
};

static PAL_Uint32 THREAD_API TestNamedSemThread(void* arg)
{
    TestNamedSemData* data = (TestNamedSemData*)arg;

    // Open or create the name semaphore:

    NamedSem sem;

    int r = NamedSem_Open(&sem, 1, SEMAPHORE_NAME, NAMEDSEM_FLAG_CREATE);

    if (r != 0)
        return NULL;

    for (unsigned int i = 0; i < data->count; i++)
    {
        NamedSem_Wait(&sem);
        size_t tmp = data->sum;
        Sleep_Milliseconds(1);
        tmp++;
        Sleep_Milliseconds(1);
        data->sum = tmp;
        Sleep_Milliseconds(1);
        NamedSem_Post(&sem, 1);
    }

    NamedSem_Close(&sem);

    return NULL;
}

static void TestNamedSem()
{
    Thread t1;
    TestNamedSemData data;

    data.sum = 0;
    data.count = 0;

    if (Thread_CreateJoinable(&t1, TestNamedSemThread, &data) != 0)
        TEST_ASSERT(false);

    Thread t2;

    if (Thread_CreateJoinable(&t2, TestNamedSemThread, &data) != 0)
        TEST_ASSERT(false);

    PAL_Uint32 arg1;
    Thread_Join(&t1, &arg1);
    Thread_Destroy(&t1);

    PAL_Uint32 arg2;
    Thread_Join(&t2, &arg2);
    Thread_Destroy(&t2);

    // Check the result:
    
    TEST_ASSERT(data.sum == 2 * data.count);
}

//==============================================================================
//
// TestShmem()
//
//==============================================================================
 
#if defined(_MSC_VER)
# define SHMEM_NAME L"/PALTestShmem"
#else
# define SHMEM_NAME "/PALTestShmem"
#endif

typedef struct _TestShmemData
{
    unsigned long sum;
}
TestShmemData;

static PAL_Uint32 THREAD_API TestShmemThread(void* arg)
{
    TestShmemData* data = (TestShmemData*)arg;
    Shmem shmem;

    // Get the size size of the shared memory:

    size_t size = 16 * Shmem_Granularity();

    // Open the shared memory segment:

    int r = Shmem_Open(&shmem, 
        SHMEM_NAME, 
        SHMEM_ACCESS_READWRITE, 
        size);

    TEST_ASSERT(r == 0);

    // Map a shared memory region:

    unsigned char* p = (unsigned char*)Shmem_Map(
        &shmem, 
        SHMEM_ACCESS_READWRITE, 
        0,
        size);

    TEST_ASSERT(p != NULL);

    // Write some data to the shared memory:

    for (size_t i = 0; i < size; i++)
    {
        TEST_ASSERT(p[i] == i % 256);
        data->sum += p[i];
        p[i] = 0xDD;
    }

    // Unmap the shared memory:

    Shmem_Unmap(&shmem, p, size);

    // Close the shared memory:

    Shmem_Close(&shmem);

    return NULL;
}

static void TestShmem()
{
    Thread t;
    Shmem shmem;

    size_t size = 16 * Shmem_Granularity();

    // Open the shared memory segment:

    int r = Shmem_Open(&shmem, 
        SHMEM_NAME, 
        SHMEM_ACCESS_READWRITE, 
        size);

    TEST_ASSERT(r == 0);

    // Map a shared memory region:

    unsigned char* p = (unsigned char*)Shmem_Map(
        &shmem, 
        SHMEM_ACCESS_READWRITE, 
        0,
        size);

    TEST_ASSERT(p != NULL);

    memset(p, 0xAA, size);

    // Write some data to the shared memory:

    unsigned long sum = 0;

    for (size_t i = 0; i < size; i++)
    {
        p[i] = i % 256;
        sum += p[i];
    }

    // Create the other thread:

    TestShmemData data;
    data.sum = 0;

    if (Thread_CreateJoinable(&t, TestShmemThread, &data) != 0)
        TEST_ASSERT(false);

    // Join with the other thread:

    PAL_Uint32 arg1;
    Thread_Join(&t, &arg1);
    Thread_Destroy(&t);

    // Compare the sums:

    TEST_ASSERT(data.sum == sum);

    // The other thread should have cleared the memory (to 0xDD's):

    for (size_t i = 0; i < size; i++)
    {
        TEST_ASSERT(p[i] == 0xDD);
    }

    // Unmap the shared memory:

    Shmem_Unmap(&shmem, p, size);

    // Close the shared memory:

    Shmem_Close(&shmem);
}

//==============================================================================
//
// TestStrings()
//
//==============================================================================

#define FUNCTION TestStringsStr
#define CHAR char
#define STRCMP Strcmp
#define STRLCPY Strlcpy
#define STRLCAT Strlcat
#define LIT(STR) STR
#include "TestStringsAux.h"
#undef FUNCTION
#undef CHAR
#undef STRCMP
#undef STRLCPY
#undef STRLCAT
#undef LIT

#define FUNCTION TestStringsWcs
#define CHAR wchar_t
#define STRCMP Wcscmp
#define STRLCPY Wcslcpy
#define STRLCAT Wcslcat
#define LIT(STR) L##STR
#include "TestStringsAux.h"
#undef FUNCTION
#undef CHAR
#undef STRCMP
#undef STRLCPY
#undef STRLCAT
#undef LIT

#define FUNCTION TestStringsTcs
#define CHAR TChar
#define STRCMP Tcscmp
#define STRLCPY Tcslcpy
#define STRLCAT Tcslcat
#define LIT(STR) PAL_T(STR)
#include "TestStringsAux.h"
#undef FUNCTION
#undef CHAR
#undef STRCMP
#undef STRLCPY
#undef STRLCAT
#undef LIT

static void TestStrings()
{
    TestStringsStr();
    TestStringsWcs();
    TestStringsTcs();
}

//==============================================================================
//
// TestDir()
//
//==============================================================================

static void TestDir()
{
    char path[PAL_MAX_PATH_SIZE]  = "/base/tests/testdir";
    Dir* dir = Dir_Open(path);
    TEST_ASSERT(dir != NULL);

    vector<string> names1;

    for (;;)
    {
        DirEnt* ent = Dir_Read(dir);
        if (!ent)
            break;

        if (strcmp(ent->name, "..") == 0 || strcmp(ent->name, ".") == 0)
            continue;

        if (strcmp(ent->name, "CVS") == 0)
            continue;

        names1.push_back(ent->name);
    }

    sort(names1.begin(), names1.end());

    vector<string> names2;
    names2.push_back("dir1");
    names2.push_back("dir2");
    names2.push_back("file1");
    names2.push_back("file2");
    names2.push_back("file3");

    TEST_ASSERT(names1.size() == 5);
    TEST_ASSERT(names2.size() == 5);
    TEST_ASSERT(names1 == names2);

    Dir_Close(dir);
}

//==============================================================================
//
// RunTests()
//
//==============================================================================

static void RunTests()
{
    printf("==== TestAtomic()\n");
    TestAtomic();

    printf("==== TestThread()\n");
    TestThread();

    printf("==== TestTLS()\n");
    TestTLS();

    printf("==== TestOnce()\n");
    TestOnce();

    printf("==== TestMutex()\n");
    TestMutex();

#if 0
    printf("==== TestCond()\n");
    TestCond();
#endif

    printf("==== TestMutex()\n");
    TestCPU();

    printf("==== TestShlib()\n");
    TestShlib();

    printf("==== TestSList()\n");
    TestSList();

    printf("==== TestSem()\n");
    TestSem();

    printf("==== TestNamedSem()\n");
    TestNamedSem();

    printf("==== TestShmem()\n");
    TestShmem();

    printf("==== TestStrings()\n");
    TestStrings();

    printf("==== TestDir()\n");
    TestDir();
}

int main()
{
    RunTests();
    return 0;
}
