#include <stdio.h>
#include "thread.h"

#if defined(PAL_HAVE_PTHREADS)
# include <sched.h>
#endif

/*
**==============================================================================
**
** Windows version
**
**==============================================================================
*/

#if defined(_MSC_VER)

int Thread_CreateDetached(
    _Out_ ThreadID* self,
    _In_ ThreadProc proc,
    _In_ void* param)
{
    HANDLE thread;

    /* Create the thread */

    thread = CreateThread(NULL, 0, proc, param, 0, &self->__impl);

    if (!thread)
    {
        return -1;
    }

    /* Close thread handle */

    CloseHandle(thread);

    return self->__impl ? 0 : -1;
}

int Thread_CreateJoinable(
    _Out_ Thread* self,
    _In_ ThreadProc proc,
    _In_ void* param)
{
    /* Create the thread */

    self->__impl = CreateThread(NULL, 0, proc, param, 0, NULL);

    return self->__impl ? 0 : -1;
}

int Thread_Join(
    _In_ Thread* self,
    _Out_ PAL_Uint32* returnValue)
{
    /* Wait for thread to exit */

    if (WaitForSingleObject(self->__impl, INFINITE) != WAIT_OBJECT_0)
    {
        return -1;
    }

    /* Get the return code */

    if (!GetExitCodeThread(self->__impl, returnValue))
    {
        return -1;
    }

    /* Close the thread heandle */

    CloseHandle(self->__impl);

    return 0;
}

void Thread_Exit(
    _In_ PAL_Uint32 returnValue)
{
    ExitThread(returnValue);
}

int Thread_Equal(
    _In_ ThreadID* thread1,
    _In_ ThreadID* thread2)
{
    return thread1->__impl == thread2->__impl;
}

ThreadID Thread_ID()
{
    ThreadID self;
    self.__impl = GetCurrentThreadId();
    return self;
}

void Thread_Yield()
{
    SwitchToThread();
}

int Thread_IDToThread(
    _In_ ThreadID* self,
    _Out_ Thread* thread)
{
    HANDLE handle = OpenThread(THREAD_ALL_ACCESS, FALSE, self->__impl);

    if (!handle)
        return -1;

    thread->__impl = handle;
    return 0;
}

#endif /* defined(_MSC_VER) */

/*
**==============================================================================
**
** POSIX threads version
**
**==============================================================================
*/

#if defined(PAL_HAVE_PTHREADS)

typedef struct _ThreadData
{
    ThreadProc proc;
    void* param;
}
ThreadData;

static void* _Wrapper(void* param) 
{
    ThreadData* data = (ThreadData*)param;
    void* result;

    result = (void*)(ptrdiff_t)(*data->proc)(data->param);

    free(data);

    return result;
}

int Thread_CreateDetached(
    _Out_ ThreadID* self,
    _In_ ThreadProc proc,
    _In_ void* param)
{
    pthread_attr_t attr;
    int r;

    /* Create the thread attributes and set the detached flag */

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    /* Create the thread */
    {
        ThreadData* data = (ThreadData*)malloc(sizeof(ThreadData));

        if (!data)
        {
            pthread_attr_destroy(&attr);
            return -1;
        }

        data->proc = proc;
        data->param = param;
        r = pthread_create(&self->__impl, &attr, _Wrapper, data);
    }

    /* Destroy thread attributes */
    pthread_attr_destroy(&attr);

    return r == 0 ? 0 : -1;
}

int Thread_CreateJoinable(
    _Out_ Thread* self,
    _In_ ThreadProc proc,
    _In_ void* param)
{
    int r;

    /* Create the thread */
    {
        ThreadData* data = (ThreadData*)malloc(sizeof(ThreadData));

        if (!data)
            return -1;

        data->proc = proc;
        data->param = param;
        r = pthread_create(&self->__impl, NULL, _Wrapper, data);
    }

    return r == 0 ? 0 : -1;
}

void Thread_Exit(
    _In_ PAL_Uint32 returnValue)
{
    pthread_exit((void*)(ptrdiff_t)returnValue);
}

int Thread_Join(
    _In_ Thread* self,
    _Out_ PAL_Uint32* returnValue)
{
    int r;
    void* returnValuePtr;
    
    r = pthread_join(self->__impl, &returnValuePtr);

    if (r == 0 && returnValue)
    {
        *returnValue = (PAL_Uint32)(ptrdiff_t)returnValuePtr;
        return 0;
    }

    return -1;
}

int Thread_Equal(
    _In_ ThreadID* thread1,
    _In_ ThreadID* thread2)
{
    return pthread_equal(thread1->__impl, thread2->__impl) == 0 ? 0 : -1;
}

ThreadID Thread_ID()
{
    ThreadID self;
    self.__impl = pthread_self();
    return self;
}

void Thread_Yield()
{
    sched_yield();
}

int Thread_IDToThread(
    _In_ ThreadID* self,
    _Out_ Thread* thread)
{
    thread->__impl = self->__impl;
    return 0;
}

#endif /* defined(PAL_HAVE_PTHREADS) */
